#!/usr/bin/python3

import sys
from pathlib import Path
from typing import Any, Final, final

import psutil
import yaml
from eth_typing import ABIElement, ABIError, HexStr
from faster_eth_abi import decode as decode_abi
from ujson import JSONDecodeError

import brownie
from brownie._c_constants import HexBytes, ujson_dump, ujson_load
from brownie._config import _get_data_folder
from brownie.convert.utils import build_function_selector, get_type_strings

# network

ERROR_SIG: Final[HexStr] = "0x08c379a0"  # type: ignore [assignment]


# error codes used in Solidity >=0.8.0
# docs.soliditylang.org/en/v0.8.0/control-structures.html#panic-via-assert-and-error-via-require
SOLIDITY_ERROR_CODES: Final = {
    1: "Failed assertion",
    17: "Integer overflow",
    18: "Division or modulo by zero",
    33: "Conversion to enum out of bounds",
    24: "Access to storage byte array that is incorrectly encoded",
    49: "Pop from empty array",
    50: "Index out of range",
    65: "Attempted to allocate too much memory",
    81: "Call to zero-initialized variable of internal function type",
}


@final
class UnknownAccount(Exception):
    pass


@final
class UndeployedLibrary(Exception):
    pass


@final
class UnsetENSName(Exception):
    pass


@final
class IncompatibleEVMVersion(Exception):
    pass


@final
class RPCProcessError(Exception):
    def __init__(self, cmd: str, uri: str) -> None:
        super().__init__(f"Unable to launch local RPC client.\nCommand: {cmd}\nURI: {uri}")


@final
class RPCConnectionError(Exception):
    def __init__(self, cmd: str, proc: psutil.Popen, uri: str) -> None:
        msg = (
            "Able to launch RPC client, but unable to connect."
            f"\n\nCommand: {cmd}\nURI: {uri}\nExit Code: {proc.poll()}"
        )
        if sys.platform != "win32":
            out = proc.stdout.read().decode().strip() or "  (Empty)"
            err = proc.stderr.read().decode().strip() or "  (Empty)"
            msg = f"{msg}\n\nStdout:\n{out}\n\nStderr:\n{err}"
        super().__init__(msg)


@final
class RPCRequestError(Exception):
    pass


@final
class MainnetUndefined(Exception):
    pass


@final
class VirtualMachineError(Exception):
    """
    Raised when a call to a contract causes an EVM exception.

    Attributes
    ----------
    message : str
        The full error message received from the RPC client.
    revert_msg : str
        The returned error string, if any.
    revert_type : str
        The error type.
    pc : int
        The program counter where the error was raised.
    txid : str
        The transaction ID that raised the error.
    """

    def __init__(self, exc: ValueError) -> None:
        self.txid: HexStr = ""  # type: ignore [assignment]
        self.source: str = ""
        self.revert_type: str = ""
        self.pc: int | None = None
        self.revert_msg: str | None = None
        self.dev_revert_msg: str | None = None

        try:
            exc = exc.args[0]
        except Exception:
            pass

        if not (isinstance(exc, dict) and "message" in exc):
            raise ValueError(str(exc)) from None

        if "data" not in exc:
            raise ValueError(exc["message"]) from None

        exc_message: str = exc["message"]
        self.message: Final[str] = exc_message.rstrip(".")

        exc_data = exc["data"]
        if isinstance(exc_data, str) and exc_data.startswith("0x"):
            self.revert_type = "revert"
            self.revert_msg = decode_typed_error(exc_data)  # type: ignore [arg-type]
            return

        try:
            txid = next(key for key in exc_data if key.startswith("0x"))
        except StopIteration:
            raise ValueError(exc["message"]) from None
        else:
            data: dict[str, Any] = exc_data[txid]
            self.revert_type = data["error"]

        self.txid = txid
        self.source = ""
        self.pc = data.get("program_counter")
        if self.pc and self.revert_type == "revert":
            self.pc -= 1

        reason = data.get("reason")
        if isinstance(reason, str) and reason.startswith("0x"):
            self.revert_msg = decode_typed_error(reason)  # type: ignore [arg-type]
        else:
            self.revert_msg = reason

        if self.pc is None:
            self.dev_revert_msg = None
        else:
            self.dev_revert_msg = brownie.project.build._get_dev_revert(self.pc)

        if self.revert_msg is None and self.revert_type in ("revert", "invalid opcode"):
            self.revert_msg = self.dev_revert_msg
        elif self.revert_msg == "Failed assertion":
            self.revert_msg = self.dev_revert_msg or self.revert_msg

    def __str__(self) -> str:
        if not hasattr(self, "revert_type"):
            return str(self.message)
        msg = self.revert_type
        if self.revert_msg:
            msg = f"{msg}: {self.revert_msg}"
        if self.source:
            msg = f"{msg}\n{self.source}"
        return str(msg)

    def _with_attr(self, **kwargs) -> "VirtualMachineError":
        for key, value in kwargs.items():
            setattr(self, key, value)
        if self.revert_msg == "Failed assertion":
            self.revert_msg = self.dev_revert_msg or self.revert_msg
        return self


@final
class TransactionError(Exception):
    pass


@final
class EventLookupError(LookupError):
    pass


@final
class NamespaceCollision(AttributeError):
    pass


# project/


@final
class ContractExists(Exception):
    pass


@final
class ContractNotFound(Exception):
    pass


@final
class ProjectAlreadyLoaded(Exception):
    pass


@final
class ProjectNotFound(Exception):
    pass


@final
class BadProjectName(Exception):
    pass


@final
class CompilerError(Exception):
    def __init__(self, e: type[psutil.Popen], compiler: str = "Compiler") -> None:
        self.compiler: Final = compiler

        err_json: dict[str, list[dict[str, str]]] = yaml.safe_load(e.stdout_data)
        err = [i.get("formattedMessage") or i["message"] for i in err_json["errors"]]
        super().__init__(f"{compiler} returned the following errors:\n\n" + "\n".join(err))


@final
class IncompatibleSolcVersion(Exception):
    pass


@final
class IncompatibleVyperVersion(Exception):
    pass


@final
class PragmaError(Exception):
    pass


@final
class InvalidManifest(Exception):
    pass


@final
class UnsupportedLanguage(Exception):
    pass


@final
class InvalidPackage(Exception):
    pass


@final
class BrownieEnvironmentError(Exception):
    pass


@final
class BrownieCompilerWarning(Warning):
    pass


class BrownieEnvironmentWarning(Warning):
    pass


@final
class InvalidArgumentWarning(BrownieEnvironmentWarning):
    pass


@final
class BrownieTestWarning(Warning):
    pass


@final
class BrownieConfigWarning(Warning):
    pass


def __get_path() -> Path:
    return _get_data_folder().joinpath("errors.json")


def parse_errors_from_abi(abi: list[ABIElement]):
    updated = False
    for item in abi:
        if item.get("type", "") == "error":
            selector = build_function_selector(item)  # type: ignore [arg-type]
            if selector in _errors:
                continue
            updated = True
            _errors[selector] = item  # type: ignore [assignment]

    if updated:
        with __get_path().open("w") as fp:
            ujson_dump(_errors, fp, sort_keys=True, indent=2)


_errors: dict[HexStr, ABIError] = {
    ERROR_SIG: {"name": "Error", "inputs": [{"name": "", "type": "string"}]}
}


try:
    with __get_path().open() as fp:
        _errors.update(ujson_load(fp))
except (FileNotFoundError, JSONDecodeError):
    pass


def decode_typed_error(data: HexStr) -> str:
    selector: HexStr = data[:10]  # type: ignore [assignment]
    if selector == "0x4e487b71":
        # special case, solidity compiler panics
        error_code = int(HexBytes(data[10:]).hex(), 16)
        return SOLIDITY_ERROR_CODES.get(error_code, f"Unknown compiler Panic: {error_code}")

    if selector not in _errors:
        return f"Unknown typed error: {data}"

    abi = _errors[selector]
    types_list = get_type_strings(abi["inputs"])
    result = decode_abi(types_list, HexBytes(data)[4:])
    return (
        result[0]
        if selector == ERROR_SIG
        else f"{abi['name']}: {', '.join(str(r) for r in result)}"
    )



#!/usr/bin/python3

import pathlib
import warnings
from typing import Any, Final

from brownie._c_constants import Path, ujson_dump, ujson_dumps, ujson_loads
from brownie._config import CONFIG
from brownie.exceptions import BrownieConfigWarning
from brownie.network.state import TxHistory
from brownie.project import get_loaded_projects
from brownie.project.build import Build
from brownie.typing import ContractName, CoverageMap
from brownie.utils import color
from brownie.utils._color import bright_green, bright_magenta, bright_red, bright_yellow

from .coverage import CoverageEval

COVERAGE_COLORS: Final[list[tuple[float, str]]] = [
    (0.8, bright_red),
    (0.9, bright_yellow),
    (1.0, bright_green),
]


def _save_coverage_report(
    build: Build, coverage_eval: CoverageEval, report_path: pathlib.Path
) -> pathlib.Path:
    # Saves a test coverage report for viewing in the GUI
    report = {
        "highlights": _get_highlights(build, coverage_eval),
        "coverage": _get_totals(build, coverage_eval),
        "sha1": {},  # TODO
    }
    report = ujson_loads(ujson_dumps(report, default=sorted))
    report_path = Path(report_path).absolute()
    if report_path.is_dir():
        report_path = report_path.joinpath("coverage.json")
    with report_path.open("w") as fp:
        ujson_dump(report, fp, sort_keys=True, indent=2)
    print(f"\nCoverage report saved at {report_path}")
    print("View the report using the Brownie GUI")
    return report_path


def _load_report_exclude_data(settings: dict[str, Any]) -> tuple[list[str], list[Any]]:
    exclude_paths: list[str] = []
    if settings["exclude_paths"]:
        exclude = settings["exclude_paths"]
        if not isinstance(exclude, list):
            exclude = [exclude]
        for glob_str in exclude:
            glob_path = Path(glob_str)
            base_path = glob_path.root if glob_path.is_absolute() else Path(".")
            try:
                exclude_paths.extend(map(Path.as_posix, base_path.glob(glob_str)))  # type: ignore [union-attr]
            except Exception:
                warnings.warn(
                    "Invalid glob pattern in config exclude settings: '{glob_str}'",
                    BrownieConfigWarning,
                )

    exclude_contracts = []
    if settings["exclude_contracts"]:
        exclude_contracts = settings["exclude_contracts"]
        if not isinstance(exclude_contracts, list):
            exclude_contracts = [exclude_contracts]

    return exclude_paths, exclude_contracts


def _build_gas_profile_output() -> list[str]:
    # Formats gas profile report that may be printed to the console
    exclude_paths, exclude_contracts = _load_report_exclude_data(CONFIG.settings["reports"])
    try:
        project = get_loaded_projects()[0]
    except IndexError:
        project = None

    gas = TxHistory().gas_profile
    sorted_gas = sorted(gas.items())
    grouped_by_contract: dict[str, dict[str, dict[str, int | str]]] = {}
    padding: dict[str, int] = {}

    lines = [""]

    only_include_project = CONFIG.settings["reports"]["only_include_project"]
    for full_name, gas_values in sorted_gas:
        contract, function = full_name.split(".", 1)

        try:
            if project._sources.get_source_path(contract) in exclude_paths:  # type: ignore [union-attr, arg-type]
                continue
        except (AttributeError, KeyError):
            # filters contracts that are not part of the project
            if only_include_project:
                continue
        if contract in exclude_contracts:
            continue

        # calculate padding to get table-like formatting
        padding["fn"] = max(padding.get("fn", 0), len(str(function)))
        for k, v in gas_values.items():
            padding[k] = max(padding.get(k, 0), len(str(v)))

        # group functions with payload by contract name
        if contract in grouped_by_contract:
            grouped_by_contract[contract][function] = gas_values  # type: ignore [assignment]
        else:
            grouped_by_contract[contract] = {function: gas_values}  # type: ignore [dict-item]

    for contract, functions in grouped_by_contract.items():
        lines.append(f"{bright_magenta}{contract}{color} <Contract>")
        sorted_functions = dict(
            sorted(functions.items(), key=lambda value: value[1]["avg"], reverse=True)
        )
        for ix, (fn_name, values) in enumerate(sorted_functions.items()):
            prefix = "\u2514\u2500" if ix == len(functions) - 1 else "\u251c\u2500"
            fn_name = fn_name.ljust(padding["fn"])
            values["avg"] = int(values["avg"])
            padded = {k: str(v).rjust(padding[k]) for k, v in values.items()}
            lines.append(
                f"   {prefix} {fn_name} -  avg: {padded['avg']}  avg (confirmed):"
                f" {padded['avg_success']}  low: {padded['low']}  high: {padded['high']}"
            )

    return lines + [""]


def _build_coverage_output(coverage_eval: CoverageEval) -> list[str]:
    # Formats a coverage evaluation report that may be printed to the console

    exclude_paths, exclude_contracts = _load_report_exclude_data(CONFIG.settings["reports"])

    projects = get_loaded_projects()
    all_totals = (
        _get_totals(project._build, coverage_eval, exclude_contracts) for project in projects
    )
    filtered = [(project, total) for project, total in zip(projects, all_totals) if total]
    lines = []

    for project, totals in filtered:

        if len(filtered) > 1:
            lines.append(f"\n======== {bright_magenta}{project._name}{color} ========")

        for contract_name in sorted(totals):
            if project._sources.get_source_path(contract_name) in exclude_paths:
                continue

            pct = _pct(
                totals[contract_name]["totals"]["statements"],
                totals[contract_name]["totals"]["branches"],
            )
            lines.append(
                f"\n  contract: {bright_magenta}{contract_name}{color}"
                f" - {_cov_color(pct)}{pct:.1%}{color}"
            )

            cov = totals[contract_name]
            branches = cov["branches"]
            results = []
            for fn_name, statement_cov in cov["statements"].items():
                branch_cov = branches.get(fn_name, (0, 0, 0))
                pct = _pct(statement_cov, branch_cov)
                results.append((fn_name, pct))

            lines.extend(
                f"    {fn_name} - {_cov_color(pct)}{pct:.1%}{color}"
                for fn_name, pct in sorted(results, key=lambda k: (-k[1], k[0]))
            )
    return lines


def _cov_color(pct: float) -> str:
    return next(i[1] for i in COVERAGE_COLORS if pct <= i[0])


def _pct(statement, branch):
    pct = statement[0] / (statement[1] or 1)
    if branch[-1]:
        pct = (pct + (branch[0] + branch[1]) / (branch[2] * 2)) / 2
    return pct


def _get_totals(
    build: Build,
    coverage_eval: CoverageEval,
    exclude_contracts: list[str] | None = None,
) -> dict[ContractName, dict[str, dict[str, dict[str, Any]]]]:
    # Returns a modified coverage eval dict showing counts and totals for each function.

    if exclude_contracts is None:
        exclude_contracts = []
    split_by_fn = _split_by_fn(build, coverage_eval)
    results: dict[ContractName, dict[str, dict[str, Any]]] = {
        i: {  # ty
            "statements": {},
            "totals": {"statements": 0, "branches": [0, 0]},
            "branches": {"true": {}, "false": {}},
        }
        for i in split_by_fn
    }
    for contract_name in split_by_fn:
        if contract_name in exclude_contracts:
            del results[contract_name]
            continue
        try:
            coverage_map = build.get(contract_name)["coverageMap"]  # type: ignore [typeddict-item]
        except KeyError:
            del results[contract_name]
            continue

        r = results[contract_name]
        r["statements"], r["totals"]["statements"] = _statement_totals(
            split_by_fn[contract_name], coverage_map["statements"], exclude_contracts
        )
        r["branches"], r["totals"]["branches"] = _branch_totals(
            split_by_fn[contract_name], coverage_map["branches"], exclude_contracts
        )

    return results


def _split_by_fn(
    build: Build,
    coverage_eval: dict[ContractName, dict[str, dict[int, set[int]]]],
) -> dict[ContractName, dict[str, dict[str, tuple[list[int], list[int], list[int]]]]]:
    # Splits a coverage eval dict so that coverage indexes are stored by function.
    results: dict[ContractName, dict[str, dict[str, Any]]] = {
        i: {"statements": {}, "branches": {"true": {}, "false": {}}} for i in coverage_eval
    }
    for name in coverage_eval:
        try:
            map_ = build.get(name)["coverageMap"]  # type: ignore [typeddict-item]
            results[name] = {k: _split(v, map_, k) for k, v in coverage_eval[name].items()}
        except KeyError:
            del results[name]
    return results


def _split(
    coverage_eval: dict[int, set[int]],
    coverage_map: CoverageMap,
    key: str,
) -> dict[str, tuple[list[int], list[int], list[int]]]:
    branches = coverage_map["branches"][key]
    statements = coverage_map["statements"][key]
    # not too sure what to call these but we don't want to getitem repeatedly
    first_eval = coverage_eval[0]
    second_eval = coverage_eval[1]
    third_eval = coverage_eval[2]
    return {
        fn: (
            [i for i in statements[fn] if int(i) in first_eval],
            [i for i in branches[fn] if int(i) in second_eval],
            [i for i in branches[fn] if int(i) in third_eval],
        )
        for fn in branches.keys() & statements.keys()
    }


def _statement_totals(
    coverage_eval: dict[str, dict[str, tuple[list[int], list[int], list[int]]]],
    coverage_map: dict[str, dict[str, dict[int, Any]]],
    exclude_contracts,
) -> tuple[dict[str, tuple[int, int]], tuple[int, int]]:
    result: dict[str, tuple[int, int]] = {}
    count, total = 0, 0
    for path, fns in coverage_eval.items():
        coverage_eval_for_path = coverage_eval[path]
        coverage_map_for_path = coverage_map[path]
        for fn in fns:
            if fn.split(".")[0] in exclude_contracts or fn not in coverage_eval_for_path:
                continue
            fn_count = len(coverage_eval_for_path[fn][0])
            fn_total = len(coverage_map_for_path[fn])
            count += fn_count
            total += fn_total
            result[fn] = fn_count, fn_total
    return result, (count, total)


def _branch_totals(
    coverage_eval: dict[str, dict[str, tuple[list[int], list[int], list[int]]]],
    coverage_map: dict[str, dict[str, dict[int, Any]]],
    exclude_contracts: list[str],
) -> tuple[dict[str, tuple[int, int, int]], tuple[int, int, int]]:
    result: dict[str, tuple[int, int, int]] = {}
    final_true = 0
    final_false = 0
    final_total = 0
    for path, fns in coverage_map.items():
        for fn in fns:
            if fn.split(".")[0] in exclude_contracts:
                continue

            if path not in coverage_eval or fn not in coverage_eval[path]:
                true, false = 0, 0
            else:
                coverage_eval_for_fn = coverage_eval[path][fn]
                true = len(coverage_eval_for_fn[2])
                false = len(coverage_eval_for_fn[1])
                final_true += true
                final_false += false

            total = len(coverage_map[path][fn])
            final_total += total

            result[fn] = (true, false, total)

    return result, (final_true, final_false, final_total)


def _get_highlights(build, coverage_eval) -> dict[str, dict[str, dict[str, list]]]:
    # Returns a highlight map formatted for display in the GUI.
    results: dict[str, dict[str, dict[str, list]]] = {"statements": {}, "branches": {}}
    for name, eval_ in coverage_eval.items():
        try:
            coverage_map = build.get(name)["coverageMap"]
        except KeyError:
            continue

        results["statements"][name] = _statement_highlights(eval_, coverage_map["statements"])
        results["branches"][name] = _branch_highlights(eval_, coverage_map["branches"])

    return results


def _statement_highlights(
    coverage_eval: dict[str, dict[int, set]],
    coverage_map: dict[str, dict[str, dict[int, Any]]],
) -> dict[str, list]:
    results: dict[str, list] = {i: [] for i in coverage_map}
    for path, fns in coverage_map.items():
        for fn in fns:
            results[path].extend(
                [*offset, _statement_color(i, coverage_eval, path), ""]
                for i, offset in coverage_map[path][fn].items()
            )
    return results


def _statement_color(
    i: int,
    coverage_eval: dict[str, dict[int, set]],
    path: str,
) -> str:
    if path in coverage_eval and int(i) in coverage_eval[path][0]:
        return "green"
    return "red"


def _branch_highlights(
    coverage_eval: dict[str, dict[int, set]],
    coverage_map: dict[str, dict[str, dict[int, Any]]],
) -> dict[str, list]:
    results: dict[str, list] = {i: [] for i in coverage_map}
    for path, fns in coverage_map.items():
        for fn in fns:
            results[path].extend(
                [*offset[:2], _branch_color(int(i), coverage_eval, path, offset[2]), ""]
                for i, offset in coverage_map[path][fn].items()
            )
    return results


def _branch_color(
    i: int,
    coverage_eval: dict[str, dict[int, set]],
    path: str,
    jump: None,
) -> str:
    if path not in coverage_eval:
        return "red"
    coverage_eval_for_path = coverage_eval[path]
    if i in coverage_eval_for_path[2]:
        if i in coverage_eval_for_path[1]:
            return "green"
        return "yellow" if jump else "orange"
    if i in coverage_eval_for_path[1]:
        return "orange" if jump else "yellow"
    return "red"



#!/usr/bin/python3

import logging
from typing import Final

import semantic_version
import vvm
import vyper
from eth_typing import ABIElement, HexStr
from packaging.version import Version as PVersion
from requests.exceptions import ConnectionError
from vyper.cli import vyper_json
from vyper.exceptions import VyperException

from brownie._c_constants import Version, deque, sha1
from brownie.exceptions import CompilerError, IncompatibleVyperVersion
from brownie.project import sources
from brownie.project.compiler.utils import VersionList, VersionSpec, expand_source_map
from brownie.project.sources import is_inside_offset
from brownie.typing import (
    Branches,
    BranchMap,
    ContractName,
    InputJsonVyper,
    Offset,
    PcList,
    PCMap,
    ProgramCounter,
    StatementMap,
    Statements,
    VyperAstJson,
    VyperAstNode,
    VyperBuildJson,
)

vvm_logger: Final = logging.getLogger("vvm")
vvm_logger.setLevel(10)
sh: Final = logging.StreamHandler()
sh.setLevel(10)
sh.setFormatter(logging.Formatter("%(message)s"))
vvm_logger.addHandler(sh)

AVAILABLE_VYPER_VERSIONS: VersionList | None = None
_active_version = Version(vyper.__version__)


EVM_VERSION_MAPPING: Final = [
    ("prague", Version("0.4.3")),
    ("cancun", Version("0.4.0")),
    ("shanghai", Version("0.3.9")),
    ("paris", Version("0.3.7")),
    ("berlin", Version("0.2.12")),
    ("istanbul", Version("0.1.0-beta.16")),
]

_get_installed_vyper_versions: Final = vvm.get_installed_vyper_versions
_get_installable_vyper_versions: Final = vvm.get_installable_vyper_versions
_vvm_set_vyper_version: Final = vvm.set_vyper_version
_vvm_install_vyper: Final = vvm.install_vyper
_vvm_compile_standard: Final = vvm.compile_standard


def get_version() -> semantic_version.Version:
    return _active_version


def set_vyper_version(version: VersionSpec) -> str:
    """Sets the vyper version. If not available it will be installed."""
    global _active_version
    if isinstance(version, str):
        version = Version(version)
    if version != Version(vyper.__version__):
        # NOTE: vvm uses `packaging.version.Version` which is not compatible with
        #       `semantic_version.Version` so we first must cast it as a string
        version_str = str(version)
        try:
            _vvm_set_vyper_version(version_str, silent=True)
        except vvm.exceptions.VyperNotInstalled:
            install_vyper(version)
            _vvm_set_vyper_version(version_str, silent=True)
    _active_version = version
    return str(_active_version)


def get_abi(contract_source: str, name: ContractName) -> dict[ContractName, list[ABIElement]]:
    """
    Given a contract source and name, returns a dict of {name: abi}

    This function is deprecated in favor of `brownie.project.compiler.get_abi`
    """
    input_json: InputJsonVyper = {  # type: ignore [typeddict-item]
        "language": "Vyper",
        "sources": {name: {"content": contract_source}},
        "settings": {"outputSelection": {"*": {"*": ["abi"]}}},
    }
    if _active_version == Version(vyper.__version__):
        try:
            compiled = vyper_json.compile_json(input_json)
        except VyperException as exc:
            raise exc.with_traceback(None)
    else:
        try:
            compiled = _vvm_compile_standard(input_json, vyper_version=str(_active_version))
        except vvm.exceptions.VyperError as exc:
            raise CompilerError(exc, "vyper")

    return {name: compiled["contracts"][name][name]["abi"]}


def _get_vyper_version_list() -> tuple[VersionList, VersionList]:
    global AVAILABLE_VYPER_VERSIONS
    installed_versions = _convert_to_semver(_get_installed_vyper_versions())
    lib_version = Version(vyper.__version__)
    if lib_version not in installed_versions:
        installed_versions.append(lib_version)
    if AVAILABLE_VYPER_VERSIONS is None:
        try:
            AVAILABLE_VYPER_VERSIONS = _convert_to_semver(_get_installable_vyper_versions())
        except ConnectionError:
            if not installed_versions:
                raise ConnectionError("Vyper not installed and cannot connect to GitHub")
            AVAILABLE_VYPER_VERSIONS = installed_versions
    return AVAILABLE_VYPER_VERSIONS, installed_versions


def install_vyper(*versions: str) -> None:
    """Installs vyper versions."""
    for version in versions:
        _vvm_install_vyper(str(version), show_progress=False)


def find_vyper_versions(
    contract_sources: dict[str, str],
    install_needed: bool = False,
    install_latest: bool = False,
    silent: bool = True,
) -> dict[str, list[str]]:
    """
    Analyzes contract pragmas and determines which vyper version(s) to use.

    Args:
        contract_sources: a dictionary in the form of {'path': "source code"}
        install_needed: if True, will install when no installed version matches
                        the contract pragma
        install_latest: if True, will install when a newer version is available
                        than the installed one
        silent: set to False to enable verbose reporting

    Returns: dictionary of {'version': ['path', 'path', ..]}
    """

    available_versions, installed_versions = _get_vyper_version_list()

    pragma_specs: dict[str, semantic_version.NpmSpec] = {}
    to_install: set[str] = set()
    new_versions: set[str] = set()

    for path, source in contract_sources.items():
        pragma_specs[path] = sources.get_vyper_pragma_spec(source, path)
        version = pragma_specs[path].select(installed_versions)

        if not version and not install_needed and not install_latest:
            raise IncompatibleVyperVersion(
                f"No installed vyper version matching '{pragma_specs[path]}' in '{path}'"
            )

        # if no installed version of vyper matches the pragma, find the latest available version
        latest = pragma_specs[path].select(available_versions)

        if not version and not latest:
            raise IncompatibleVyperVersion(
                f"No installable vyper version matching '{pragma_specs[path]}' in '{path}'"
            )

        if not version or (install_latest and latest > version):
            to_install.add(str(latest))
        elif latest and latest > version:
            new_versions.add(str(version))

    # install new versions if needed
    if to_install:
        install_vyper(*to_install)
        _, installed_versions = _get_vyper_version_list()
    elif new_versions and not silent:
        print(
            f"New compatible vyper version{'s' if len(new_versions) > 1 else ''}"
            f" available: {', '.join(new_versions)}"
        )

    # organize source paths by latest available vyper version
    compiler_versions: dict[str, list[str]] = {}
    for path, spec in pragma_specs.items():
        version = spec.select(installed_versions)
        compiler_versions.setdefault(str(version), []).append(path)

    return compiler_versions


def find_best_vyper_version(
    contract_sources: dict[str, str],
    install_needed: bool = False,
    install_latest: bool = False,
    silent: bool = True,
) -> str:
    """
    Analyze contract pragma and find the best compatible version across multiple sources.

    Args:
        contract_sources: a dictionary in the form of {'path': "source code"}
        install_needed: if True, will install when no installed version matches
                        the contract pragma
        install_latest: if True, will install when a newer version is available
                        than the installed one
        silent: set to False to enable verbose reporting

    Returns: version string
    """

    available_versions, installed_versions = _get_vyper_version_list()

    for path, source in contract_sources.items():

        pragma_spec = sources.get_vyper_pragma_spec(source, path)
        installed_versions = [i for i in installed_versions if i in pragma_spec]
        available_versions = [i for i in available_versions if i in pragma_spec]

    if not available_versions:
        raise IncompatibleVyperVersion("No installable vyper version compatible across all sources")

    if not installed_versions and not install_needed and not install_latest:
        raise IncompatibleVyperVersion("No installed vyper version compatible across all sources")

    if max(available_versions) > max(installed_versions, default=Version("0.0.0")):
        if install_latest or (install_needed and not installed_versions):
            install_vyper(max(available_versions))
            return str(max(available_versions))
        if not silent:
            print(f"New compatible vyper version available: {max(available_versions)}")

    return str(max(installed_versions))


def compile_from_input_json(
    input_json: InputJsonVyper, silent: bool = True, allow_paths: str | None = None
) -> dict:
    """
    Compiles contracts from a standard input json.

    Args:
        input_json: vyper input json
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns: standard compiler output json
    """

    version = get_version()
    if not silent:
        print("Compiling contracts...")
        print(f"  Vyper version: {version}")
    if version < Version("0.1.0-beta.17"):
        outputs = input_json["settings"]["outputSelection"]["*"]["*"]
        outputs.remove("userdoc")
        outputs.remove("devdoc")
    if version == Version(vyper.__version__):
        try:
            return vyper_json.compile_json(input_json)
        except VyperException as exc:
            raise exc.with_traceback(None)
    else:
        try:
            # NOTE: vvm uses `packaging.version.Version` which is not compatible with
            #       `semantic_version.Version` so we first must cast it as a string
            version = str(version)
            return _vvm_compile_standard(input_json, base_path=allow_paths, vyper_version=version)
        except vvm.exceptions.VyperError as exc:
            raise CompilerError(exc, "vyper")


def _get_unique_build_json(
    output_evm: dict,
    path_str: str,
    contract_name: ContractName,
    ast_json: dict | list,
    offset: Offset,
) -> VyperBuildJson:

    ast: list = ast_json["body"] if isinstance(ast_json, dict) else ast_json
    deployed_bytecode: dict = output_evm["deployedBytecode"]
    pc_map, statement_map, branch_map = _generate_coverage_data(
        deployed_bytecode["sourceMap"],
        deployed_bytecode["opcodes"],
        contract_name,
        ast,
    )
    bytecode_json: dict = output_evm["bytecode"]
    bytecode: HexStr = bytecode_json["object"]
    # This is not a complete VyperBuildJson object but that's okay for now
    # TODO: make a partial VyperBuildJson subclass for this dtype
    return {  # type: ignore [typeddict-item]
        "allSourcePaths": {"0": path_str},
        "bytecode": bytecode,
        "bytecodeSha1": sha1(bytecode.encode()).hexdigest(),  # type: ignore [typeddict-item]
        "coverageMap": {"statements": statement_map, "branches": branch_map},
        "dependencies": _get_dependencies(ast),
        "offset": offset,
        "pcMap": pc_map,
        "type": "contract",
    }


def _get_dependencies(ast_json: list[dict]) -> list[ContractName]:
    return sorted(
        {
            i["name"].split(".")[-1]
            for i in ast_json
            if i["ast_type"] == "Import"
            or (i["ast_type"] == "ImportFrom" and i["module"] != "vyper.interfaces")
        }
    )


def _is_revert_jump(pc_list: PcList, revert_pc: int) -> bool:
    return pc_list[-1]["op"] == "JUMPI" and int(pc_list[-2].get("value", "0"), 16) == revert_pc


def _generate_coverage_data(
    source_map_str: str,
    opcodes_str: str,
    contract_name: ContractName,
    ast_json: VyperAstJson,
) -> tuple[PCMap, StatementMap, BranchMap]:
    if not opcodes_str:
        return PCMap({}), {}, {}
        return {}, {}, {}

    source_map = deque(expand_source_map(source_map_str))
    opcodes = deque(opcodes_str.split(" "))

    fn_nodes = [i for i in ast_json if i["ast_type"] == "FunctionDef"]
    fn_offsets = {i["name"]: _convert_src(i["src"]) for i in fn_nodes}
    stmt_nodes = {_convert_src(i["src"]) for i in _get_statement_nodes(fn_nodes)}

    statement_map: Statements = {}
    branch_map: Branches = {}

    pc_list: PcList = []
    count, pc = 0, 0

    revert_pc = -1
    if opcodes[-5] == "JUMPDEST" and opcodes[-1] == "REVERT":
        # starting in vyper 0.2.14, reverts without a reason string are optimized
        # with a jump to the end of the bytecode. if the bytecode ends with this
        # pattern, we set `revert_pc` as the program counter of the jumpdest so
        # we can identify these optimizer reverts within traces.
        revert_pc = len(opcodes) + sum(int(i[4:]) - 1 for i in opcodes if i.startswith("PUSH")) - 5

    while opcodes and source_map:

        # format of source is [start, stop, contract_id, jump code]
        start, stop, _, jump_code = source_map.popleft()

        op = opcodes.popleft()
        this: ProgramCounter = {"op": op, "pc": pc}  # type: ignore [typeddict-item]
        pc_list.append(this)

        if jump_code != "-":
            this["jump"] = jump_code

        pc += 1
        if opcodes and opcodes[0][:2] == "0x":
            this["value"] = opcodes.popleft()
            pc += int(op[4:])

        # set source offset (-1 means none)
        if start == -1:
            if (len(pc_list) > 6 and pc_list[-7]["op"] == "CALLVALUE" and op == "REVERT") or (
                len(pc_list) > 2
                and pc_list[-3]["op"] == "CALLVALUE"
                and _is_revert_jump(pc_list[-2:], revert_pc)
            ):
                # special case - initial nonpayable check on vyper >=0.2.5
                this["dev"] = "Cannot send ether to nonpayable function"
                # hackiness to prevent the source highlight from showing the entire contract
                if op == "REVERT":
                    # for REVERT, apply to the previous opcode
                    pc_list[-2].update(path="0", offset=(0, 0))  # type: ignore [call-arg]
                else:
                    # for JUMPI we need the mapping on the actual opcode
                    this.update(path="0", offset=(0, 0))  # type: ignore [call-arg]
            continue

        offset: Offset = (start, start + stop)  # type: ignore [assignment]
        this["path"] = "0"
        this["offset"] = offset

        try:
            if "offset" in pc_list[-2] and offset == pc_list[-2]["offset"]:
                this_fn = this["fn"] = pc_list[-2]["fn"]
            else:
                # statement coverage
                fn = next(k for k, v in fn_offsets.items() if is_inside_offset(offset, v))
                this_fn = this["fn"] = f"{contract_name}.{fn}"
                stmt_offset = next(i for i in stmt_nodes if is_inside_offset(offset, i))
                stmt_nodes.remove(stmt_offset)
                statement_map.setdefault(this_fn, {})[count] = stmt_offset
                this["statement"] = count
                count += 1
        except (KeyError, IndexError, StopIteration):
            pass

        if op not in ("JUMPI", "REVERT"):
            continue

        node = _find_node_by_offset(ast_json, offset)
        if node is None:
            continue

        node_ast_type = node["ast_type"]
        if op == "REVERT" or _is_revert_jump(pc_list[-2:], revert_pc):
            # custom revert error strings
            if node_ast_type == "FunctionDef":
                if (op == "REVERT" and pc_list[-7]["op"] == "CALLVALUE") or (
                    op == "JUMPI" and pc_list[-3]["op"] == "CALLVALUE"
                ):
                    this["dev"] = "Cannot send ether to nonpayable function"
            elif node_ast_type == "Subscript":
                this["dev"] = "Index out of range"
            elif node_ast_type in ("AugAssign", "BinOp"):
                node_op: VyperAstNode = node["op"]
                node_op_ast_type = node_op["ast_type"]
                if node_op_ast_type == "Sub":
                    this["dev"] = "Integer underflow"
                elif node_op_ast_type == "Div":
                    this["dev"] = "Division by zero"
                elif node_op_ast_type == "Mod":
                    this["dev"] = "Modulo by zero"
                else:
                    this["dev"] = "Integer overflow"
            continue

        if node_ast_type in ("Assert", "If") or (
            node_ast_type == "Expr"
            and node["value"].get("func", {}).get("id") == "assert_modifiable"
        ):
            # branch coverage
            this["branch"] = count
            this_fn = this["fn"]
            branch_map.setdefault(this_fn, {})  # type: ignore [arg-type]
            if node_ast_type == "If":
                branch_map[this_fn][count] = _convert_src(node["test"]["src"]) + (False,)  # type: ignore [index]
            else:
                branch_map[this_fn][count] = offset + (True,)  # type: ignore [index]
            count += 1

    first = pc_list[0]
    first["path"] = "0"
    first["offset"] = (0, _convert_src(ast_json[-1]["src"])[1])
    if revert_pc != -1:
        this["optimizer_revert"] = True

    pc_map = PCMap({i.pop("pc"): i for i in pc_list})

    return pc_map, {"0": statement_map}, {"0": branch_map}


def _convert_src(src: str) -> Offset:
    if src is None:
        return -1, -1
    split = src.split(":")[:2]
    start = int(split[0])
    stop = start + int(split[1])
    return start, stop


def _find_node_by_offset(ast_json: VyperAstJson, offset: Offset) -> VyperAstNode | None:
    for node in ast_json:
        converted_src = _convert_src(node["src"])
        if is_inside_offset(offset, converted_src):
            if converted_src == offset:
                return node
            node_list: VyperAstJson = [
                i for i in node.values() if isinstance(i, dict) and "ast_type" in i  # type: ignore [misc]
            ]
            for v in node.values():
                if isinstance(v, list):
                    node_list.extend(v)
            if node_list:
                result = _find_node_by_offset(node_list, offset)
            else:
                result = _find_node_by_offset(ast_json[ast_json.index(node) + 1 :], offset)
            if result is not None:
                return result
    return None


def _get_statement_nodes(ast_json: VyperAstJson) -> VyperAstJson:
    stmt_nodes = []
    for node in ast_json:
        if children := [x for v in node.values() if isinstance(v, list) for x in v]:
            stmt_nodes += _get_statement_nodes(children)
        else:
            stmt_nodes.append(node)
    return stmt_nodes


def _convert_to_semver(versions: list[PVersion]) -> VersionList:
    """
    Converts a list of `packaging.version.Version` objects to a list of
    `semantic_version.Version` objects.

    vvm 0.2.0 switched to packaging.version but we are not ready to
    migrate brownie off of semantic-version.

    This function serves as a stopgap.
    """
    return [  # type: ignore [return-value]
        Version(
            major=version.major,
            minor=version.minor,
            patch=version.micro,
            prerelease="".join(str(x) for x in version.pre) if version.pre else None,
        )
        for version in versions
    ]



#!/usr/bin/python3

import sys
import warnings
from subprocess import DEVNULL, PIPE
from typing import Any

import psutil
from requests.exceptions import ConnectionError as RequestsConnectionError

from brownie.exceptions import InvalidArgumentWarning, RPCRequestError
from brownie.network.web3 import web3

CLI_FLAGS = {
    "port": "--port",
    "host": "--host",
    "evm_version": "--hardfork",
    "fork": "--fork-url",
    "fork_block": "--fork-block-number",
    "chain_id": "--chain-id",
    "default_balance": "--balance",
    "accounts": "--accounts",
    "mnemonic": "--mnemonic",
    "block_time": "--block-time",
    "base_fee": "--block-base-fee-per-gas",
    "gas_price": "--gas-price",
    "gas_limit": "--gas-limit",
}


def launch(cmd: str, **kwargs: Any) -> None:
    """Launches the RPC client.

    Args:
        cmd: command string to execute as subprocess"""
    if sys.platform == "win32" and not cmd.split(" ")[0].endswith(".cmd"):
        if " " in cmd:
            cmd = cmd.replace(" ", ".cmd ", 1)
        else:
            cmd += ".cmd"
    cmd_list = cmd.split(" ")
    for key, value in [(k, v) for k, v in kwargs.items() if v]:
        try:
            cmd_list.extend([CLI_FLAGS[key], str(value)])
        except KeyError:
            warnings.warn(
                f"Ignoring invalid commandline setting for anvil: "
                f'"{key}" with value "{value}".',
                InvalidArgumentWarning,
            )
    print(f"\nLaunching '{' '.join(cmd_list)}'...")
    out = DEVNULL if sys.platform == "win32" else PIPE

    return psutil.Popen(cmd_list, stdin=DEVNULL, stdout=out, stderr=out)


def on_connection() -> None:
    # set gas limit to the same as the forked network
    gas_limit = web3.eth.get_block("latest").gasLimit
    web3.provider.make_request("evm_setBlockGasLimit", [hex(gas_limit)])


def _request(method: str, args: list) -> int:
    try:
        response = web3.provider.make_request(method, args)
        if "result" in response:
            return response["result"]
    except (AttributeError, RequestsConnectionError):
        raise RPCRequestError("Web3 is not connected.")
    raise RPCRequestError(response["error"]["message"])


def sleep(seconds: int) -> int:
    _request("evm_increaseTime", [hex(seconds)])
    return seconds


def mine(blocks: int | None = None) -> None:
    _request("evm_mine", blocks)


def snapshot() -> int:
    return _request("evm_snapshot", [])


def revert(snapshot_id: int) -> None:
    _request("evm_revert", [snapshot_id])


def unlock_account(address: str) -> None:
    web3.provider.make_request("anvil_impersonateAccount", [address])



#!/usr/bin/python3

import warnings
from typing import Final

from brownie import project
from brownie._config import CONFIG
from brownie.convert import Wei
from brownie.exceptions import BrownieEnvironmentWarning

from .account import Accounts
from .gas.bases import GasABC
from .rpc import Rpc
from .state import Chain, _notify_registry
from .web3 import web3

chain: Final = Chain()
rpc: Final = Rpc()


def connect(network: str | None = None, launch_rpc: bool = True) -> None:
    """Connects to the network.

    Args:
        network: string of the name of the network to connect to

    Network information is retrieved from brownie-config.json"""
    if is_connected():
        raise ConnectionError(f"Already connected to network '{CONFIG.active_network['id']}'")
    try:
        active = CONFIG.set_active_network(network)
        host = active["host"]

        if ":" not in host.split("//", maxsplit=1)[-1]:
            try:
                host += f":{active['cmd_settings']['port']}"
            except KeyError:
                pass

        web3.connect(host, active.get("timeout", 30))
        if CONFIG.network_type == "development" and launch_rpc and not rpc.is_active():
            if is_connected():
                if web3.eth.block_number != 0:
                    warnings.warn(
                        f"Development network has a block height of {web3.eth.block_number}",
                        BrownieEnvironmentWarning,
                    )
                rpc.attach(host)
            else:
                rpc.launch(active["cmd"], **active["cmd_settings"])
        else:
            Accounts()._reset()
        if CONFIG.network_type == "live" or CONFIG.settings["dev_deployment_artifacts"]:
            for p in project.get_loaded_projects():
                p._load_deployments()

    except Exception:
        CONFIG.clear_active()
        web3.disconnect()
        raise


def disconnect(kill_rpc: bool = True) -> None:
    """Disconnects from the network."""
    if not is_connected():
        raise ConnectionError("Not connected to any network")
    CONFIG.clear_active()
    if kill_rpc and rpc.is_active() and rpc.is_child():
        rpc.kill()
    web3.disconnect()
    _notify_registry(0)


def show_active() -> str | None:
    """Returns the name of the currently active network"""
    return CONFIG.active_network["id"] if web3.provider else None


def is_connected() -> bool:
    """Returns a bool indicating if the Web3 object is currently connected"""
    return web3.isConnected()


def gas_limit(*args: int | str | bool | None) -> int | bool:
    """Gets and optionally sets the default gas limit.

    * If an integer value is given, this will be the default gas limit.
    * If set to 'auto', the gas limit is determined automatically."""

    if not is_connected():
        raise ConnectionError("Not connected to any network")
    if args:
        if args[0] in (None, False, True, "auto"):
            CONFIG.active_network["settings"]["gas_limit"] = False
        else:
            try:
                limit: int = int(args[0])
            except ValueError:
                raise TypeError(f"Invalid gas limit '{args[0]}'")
            if limit < 21000:
                raise ValueError("Minimum gas limit is 21000")
            CONFIG.active_network["settings"]["gas_limit"] = limit
    return CONFIG.active_network["settings"]["gas_limit"]


def gas_price(*args: int | str | bool | None) -> int | bool:
    """Gets and optionally sets the default gas price.

    * If an integer value is given, this will be the default gas price.
    * If set to 'auto', the gas price is determined automatically."""

    if not is_connected():
        raise ConnectionError("Not connected to any network")
    if args:
        if isinstance(args[0], GasABC):
            CONFIG.active_network["settings"]["gas_price"] = args[0]
        elif args[0] in (None, False, True, "auto"):
            CONFIG.active_network["settings"]["gas_price"] = False
        else:
            try:
                price = Wei(args[0])
            except ValueError:
                raise TypeError(f"Invalid gas price '{args[0]}'")
            CONFIG.active_network["settings"]["gas_price"] = price
    return CONFIG.active_network["settings"]["gas_price"]


def gas_buffer(*args: float | None) -> float | None:
    if not is_connected():
        raise ConnectionError("Not connected to any network")
    if args:
        if args[0] is None:
            CONFIG.active_network["settings"]["gas_buffer"] = 1
        elif isinstance(args[0], (float, int)):
            CONFIG.active_network["settings"]["gas_buffer"] = args[0]
        else:
            raise TypeError("Invalid gas buffer - must be given as a float, int or None")
    return CONFIG.active_network["settings"]["gas_buffer"]


def max_fee(*args: int | str | bool | None) -> int | bool:
    """
    Gets and optionally sets the default max fee per gas.

    * If a Wei value is given, this will be the default max fee.
    * If set to None or False, transactions will default to using gas price.
    """
    if not is_connected():
        raise ConnectionError("Not connected to any network")
    if args:
        if args[0] in (None, False):
            CONFIG.active_network["settings"]["max_fee"] = None
        else:
            try:
                price = Wei(args[0])
            except ValueError:
                raise TypeError(f"Invalid max fee '{args[0]}'")
            CONFIG.active_network["settings"]["max_fee"] = price
    return CONFIG.active_network["settings"]["max_fee"]


def priority_fee(*args: int | str | bool | None) -> int | bool:
    """
    Gets and optionally sets the default max priority fee per gas.

    * If set to 'auto', the fee is set using `eth_maxPriorityFeePerGas`.
    * If a Wei value is given, this will be the default max fee.
    * If set to None or False, transactions will default to using gas price.
    """
    if not is_connected():
        raise ConnectionError("Not connected to any network")
    if args:
        if args[0] in (None, False):
            CONFIG.active_network["settings"]["priority_fee"] = None
        elif args[0] == "auto":
            CONFIG.active_network["settings"]["priority_fee"] = "auto"
        else:
            try:
                price = Wei(args[0])
            except ValueError:
                raise TypeError(f"Invalid priority fee '{args[0]}'")
            CONFIG.active_network["settings"]["priority_fee"] = price
    return CONFIG.active_network["settings"]["priority_fee"]



#!/usr/bin/python3
import os
import platform
import sys

from setuptools import find_packages, setup

with open("README.md") as fh:
    long_description = fh.read()

if os.environ.get("BROWNIE_LIB", "0") == "1":
    if sys.platform == "windows":
        requirements_filename = "requirements-windows.in"
    else:
        requirements_filename = "requirements.in"
elif sys.platform == "windows":
    requirements_filename = "requirements-windows.txt"
else:
    requirements_filename = "requirements.txt"

with open(requirements_filename) as f:
    requirements = list(map(str.strip, f.read().split("\n")))[:-1]

try:
    from mypyc.build import mypycify
except ImportError:
    skip_mypyc = True
else:
    # We only compile this library for CPython, other implementations will use it as normal interpreted python code
    # CPython users can also set the BROWNIE_NOCOMPILE env at install time to run brownie in interpreted python mode
    skip_mypyc = (
        os.environ.get("BROWNIE_NOCOMPILE")
        or platform.python_implementation() != "CPython"
        or any(
            cmd in sys.argv
            for cmd in ("sdist", "egg_info", "--name", "--version", "--help", "--help-commands")
        )
    )
if skip_mypyc:
    ext_modules = []
else:
    ext_modules = mypycify(
        [
            "brownie/_c_constants.py",
            "brownie/_cli",
            "brownie/_config.py",
            "brownie/_expansion.py",
            "brownie/convert",
            "brownie/network/__init__.py",
            "brownie/network/alert.py",
            "brownie/network/event.py",
            "brownie/network/middlewares/__init__.py",
            "brownie/network/middlewares/caching.py",
            "brownie/network/middlewares/catch_tx_revert.py",
            "brownie/network/middlewares/ganache7.py",
            "brownie/network/middlewares/geth_poa.py",
            "brownie/network/middlewares/hardhat.py",
            "brownie/network/state.py",
            "brownie/project",
            "brownie/test/coverage.py",
            "brownie/test/managers/utils.py",
            "brownie/test/output.py",
            "brownie/test/stateful.py",
            "brownie/typing.py",
            "brownie/utils/__init__.py",
            "brownie/utils/_color.py",
            "brownie/utils/output.py",
            "brownie/utils/sql.py",
            "brownie/utils/toposort.py",
            # "--strict",
            "--pretty",
            "--check-untyped-defs",
        ],
        group_name="eth_brownie",
        strict_dunder_typing=True,
    )


setup(
    name="eth-brownie",
    packages=find_packages(),
    version="1.22.0.dev1",  # don't change this manually, use bumpversion instead
    license="MIT",
    description="A Python framework for Ethereum smart contract deployment, testing and interaction.",  # noqa: E501
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ben Hauser",
    author_email="ben@hauser.id",
    url="https://github.com/eth-brownie/brownie",
    keywords=["brownie"],
    install_requires=requirements,
    entry_points={
        "console_scripts": ["brownie=brownie._cli.__main__:main"],
        "pytest11": ["pytest-brownie=brownie.test.plugin"],
    },
    package_data={"brownie": ["py.typed"]},
    include_package_data=True,
    ext_modules=ext_modules,
    python_requires=">=3.10,<4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
)



#!/usr/bin/python3

import sys
import warnings
from pathlib import Path
from subprocess import DEVNULL, PIPE
from typing import Any

import psutil
from requests.exceptions import ConnectionError as RequestsConnectionError

from brownie.exceptions import InvalidArgumentWarning, RPCRequestError
from brownie.network.web3 import web3

CLI_FLAGS = {
    "port": "--port",
    "host": "--hostname",
    "fork": "--fork",
    "fork_block": "--fork-block-number",
}
IGNORED_SETTINGS = ["chain_id"]

HARDHAT_CONFIG = """
// autogenerated by brownie
// do not modify the existing settings
module.exports = {
    networks: {
        hardhat: {
            hardfork: "london",
            // base fee of 0 allows use of 0 gas price when testing
            initialBaseFeePerGas: 0,
            // brownie expects calls and transactions to throw on revert
            throwOnTransactionFailures: true,
            throwOnCallFailures: true
       }
    }
}"""


def launch(cmd: str, **kwargs: Any) -> None:
    """Launches the RPC client.

    Args:
        cmd: command string to execute as subprocess"""
    if sys.platform == "win32" and not cmd.split(" ")[0].endswith(".cmd"):
        if " " in cmd:
            cmd = cmd.replace(" ", ".cmd ", 1)
        else:
            cmd += ".cmd"
    cmd_list = cmd.split(" ")
    for key, value in [(k, v) for k, v in kwargs.items() if v and k not in IGNORED_SETTINGS]:
        try:
            cmd_list.extend([CLI_FLAGS[key], str(value)])
        except KeyError:
            warnings.warn(
                f"Ignoring invalid commandline setting for hardhat: "
                f'"{key}" with value "{value}".',
                InvalidArgumentWarning,
            )
    print(f"\nLaunching '{' '.join(cmd_list)}'...")
    out = DEVNULL if sys.platform == "win32" else PIPE

    # check parent folders for existence of a hardhat config, so this folder is
    # considered a hardhat project. if none is found, create one.
    config_exists = False
    for path in Path("hardhat.config.js").absolute().parents:
        if path.joinpath("hardhat.config.js").exists():
            config_exists = True
            break
    if not config_exists:
        with Path("hardhat.config.js").open("w") as fp:
            fp.write(HARDHAT_CONFIG)

    return psutil.Popen(cmd_list, stdin=DEVNULL, stdout=out, stderr=out)


def on_connection() -> None:
    gas_limit = web3.eth.get_block("latest").gasLimit
    web3.provider.make_request("evm_setBlockGasLimit", [hex(gas_limit)])


def _request(method: str, args: list) -> int:
    try:
        response = web3.provider.make_request(method, args)
        if "result" in response:
            return response["result"]
    except (AttributeError, RequestsConnectionError):
        raise RPCRequestError("Web3 is not connected.")
    raise RPCRequestError(response["error"]["message"])


def sleep(seconds: int) -> int:
    return _request("evm_increaseTime", [seconds])


def mine(timestamp: int | None = None) -> None:
    params = [timestamp] if timestamp else []
    _request("evm_mine", params)


def snapshot() -> int:
    return _request("evm_snapshot", [])


def revert(snapshot_id: int) -> None:
    _request("evm_revert", [snapshot_id])


def unlock_account(address: str) -> None:
    web3.provider.make_request("hardhat_impersonateAccount", [address])



#!/usr/bin/python3

import atexit
import inspect
import platform
import socket
import time
import warnings
from collections.abc import Callable
from typing import Any
from urllib.parse import urlparse

import psutil

from brownie._singleton import _Singleton
from brownie.exceptions import RPCConnectionError, RPCProcessError
from brownie.network.state import Chain
from brownie.network.web3 import web3

from . import anvil, ganache, geth, hardhat

chain = Chain()

ATTACH_BACKENDS = {"ethereumjs testrpc": ganache, "geth": geth, "hardhat": hardhat}

LAUNCH_BACKENDS = {
    "anvil": anvil,
    "ganache": ganache,
    "ethnode": geth,
    "geth": geth,
    "npx hardhat": hardhat,
}


def internal(fn: Callable) -> Callable:
    """
    Decorator that warns when a user calls directly to Rpc methods.
    """

    def wrapped(*args: Any, **kwargs: Any) -> Callable:
        if inspect.stack()[1].frame.f_locals.get("self", None) != chain:
            warnings.warn(
                f"rpc.{fn.__name__} should not be called directly, use chain.{fn.__name__} instead",
                FutureWarning,
                stacklevel=2,
            )
        return fn(*args, **kwargs)

    return wrapped


class Rpc(metaclass=_Singleton):
    def __init__(self) -> None:
        self.process: psutil.Popen | psutil.Process = None
        self.backend: Any = ganache
        atexit.register(self._at_exit)

    def _at_exit(self) -> None:
        if not self.is_active():
            return
        if self.process.parent() == psutil.Process():
            if getattr(self.process, "stdout", None) is not None:
                self.process.stdout.close()
            if getattr(self.process, "stderr", None) is not None:
                self.process.stderr.close()
            self.kill(False)

    def launch(self, cmd: str, **kwargs: Any) -> None:
        if self.is_active():
            raise SystemError("RPC is already active.")

        for key, module in LAUNCH_BACKENDS.items():
            if cmd.lower().startswith(key):
                self.backend = module
                break

        self.process = self.backend.launch(cmd, **kwargs)

        # check that web3 can connect
        if not web3.provider:
            chain._network_disconnected()
            return
        uri = web3.provider.endpoint_uri if web3.provider else None
        for i in range(100):
            if web3.isConnected():
                web3.reset_middlewares()
                self.backend.on_connection()
                chain._network_connected()
                return
            time.sleep(0.1)
            if isinstance(self.process, psutil.Popen):
                self.process.poll()
            if not self.process.is_running():
                self.kill(False)
                raise RPCProcessError(cmd, uri)
        self.kill(False)
        raise RPCConnectionError(cmd, self.process, uri)

    def attach(self, laddr: str | tuple) -> None:
        """Attaches to an already running RPC client subprocess.

        Args:
            laddr: Address that the client is listening at. Can be supplied as a
                   string "http://127.0.0.1:8545" or tuple ("127.0.0.1", 8545)"""
        if self.is_active():
            raise SystemError("RPC is already active.")
        if isinstance(laddr, str):
            o = urlparse(laddr)
            if not o.port:
                raise ValueError("No RPC port given")
            laddr = (o.hostname, o.port)

        ip = socket.gethostbyname(laddr[0])
        resolved_addr = (ip, laddr[1])
        pid = self._find_rpc_process_pid(resolved_addr)

        print(f"Attached to local RPC client listening at '{laddr[0]}:{laddr[1]}'...")
        self.process = psutil.Process(pid)

        for key, module in ATTACH_BACKENDS.items():
            if web3.client_version.lower().startswith(key):
                self.backend = module
                break

        web3.reset_middlewares()
        self.backend.on_connection()
        chain._network_connected()

    def kill(self, exc: bool = True) -> None:
        """Terminates the RPC process and all children with SIGKILL.

        Args:
            exc: if True, raises SystemError if subprocess is not active."""
        if not self.is_active():
            if not exc:
                return
            raise SystemError("RPC is not active.")

        try:
            print("Terminating local RPC client...")
        except ValueError:
            pass
        for child in self.process.children(recursive=True):
            try:
                child.kill()
            except psutil.NoSuchProcess:
                pass
        self.process.kill()
        self.process.wait()
        chain._network_disconnected()

    def is_active(self) -> bool:
        """Returns True if Rpc client is currently active."""
        if not self.process:
            return False
        if isinstance(self.process, psutil.Popen):
            self.process.poll()
        return self.process.is_running()

    def is_child(self) -> bool:
        """Returns True if the Rpc client is active and was launched by Brownie."""
        if not self.is_active():
            return False
        return self.process.parent() == psutil.Process()

    @internal
    def sleep(self, seconds: int) -> int:
        return self.backend.sleep(seconds)

    @internal
    def mine(self, timestamp: int = None) -> int:
        self.backend.mine(timestamp)
        return web3.eth.block_number

    @internal
    def snapshot(self) -> int:
        return self.backend.snapshot()

    @internal
    def revert(self, snapshot_id: int) -> int:
        self.backend.revert(snapshot_id)
        return web3.eth.block_number

    def unlock_account(self, address: str) -> None:
        self.backend.unlock_account(address)

    def _find_rpc_process_pid(self, laddr: tuple) -> int:
        try:
            # default case with an already running local RPC process
            return self._get_pid_from_connections(laddr)
        except ProcessLookupError:
            # if no local RPC process could be found we can try to find a dockerized one
            if platform.system() == "Darwin":
                return self._get_pid_from_docker_backend()
            else:
                return self._get_pid_from_net_connections(laddr)

    def _check_proc_connections(self, proc: psutil.Process, laddr: tuple) -> bool:
        try:
            return laddr in [i.laddr for i in proc.connections()]
        except psutil.AccessDenied:
            return False
        except psutil.ZombieProcess:
            return False
        except psutil.NoSuchProcess:
            return False

    def _check_net_connections(self, connection: Any, laddr: tuple) -> bool:
        if connection.pid is None:
            return False
        if connection.laddr == laddr:
            return True
        elif connection.raddr == laddr:
            return True
        else:
            return False

    def _get_pid_from_connections(self, laddr: tuple) -> int:
        try:
            proc = next(i for i in psutil.process_iter() if self._check_proc_connections(i, laddr))
            return self._get_proc_pid(proc)
        except StopIteration:
            raise ProcessLookupError(
                "Could not attach to RPC process by querying 'proc.connections()'"
            ) from None

    def _get_pid_from_net_connections(self, laddr: tuple) -> int:
        try:
            proc = next(
                i
                for i in psutil.net_connections(kind="tcp")
                if self._check_net_connections(i, laddr)
            )
            return self._get_proc_pid(proc)
        except StopIteration:
            raise ProcessLookupError(
                "Could not attach to RPC process by querying 'proc.net_connections()'"
            ) from None

    def _get_pid_from_docker_backend(self) -> int:
        # OSX workaround for https://github.com/giampaolo/psutil/issues/1219
        proc = self._find_proc_by_name("com.docker.backend")
        return self._get_proc_pid(proc)

    def _get_proc_pid(self, proc: psutil.Process) -> int:
        if proc:
            return proc.pid
        else:
            raise ProcessLookupError(
                "Could not attach to RPC process. If this issue persists, try killing "
                "the RPC and let Brownie launch it as a child process."
            ) from None

    def _find_proc_by_name(self, process_name: str) -> psutil.Process:
        for proc in psutil.process_iter():
            if process_name.lower() in proc.name().lower():
                return proc



#!/usr/bin/python3


class _Singleton(type):

    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        # NOTE counterintuitively, when you supply type hints for a metaclass call method,
        #      it breaks the standard typing for the created instances. Do not add types here.
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



#!/usr/bin/python3

import gc
import threading
import time
import weakref
from collections.abc import Callable, Iterator
from pathlib import Path
from sqlite3 import OperationalError
from typing import TYPE_CHECKING, Any, Final, Union, final

from eth_typing import BlockNumber, ChecksumAddress, HexAddress, HexStr
from web3.datastructures import AttributeDict
from web3.types import BlockData

import brownie.network.rpc as rpc
from brownie._c_constants import sha1
from brownie._config import CONFIG, _get_data_folder
from brownie._singleton import _Singleton
from brownie.convert import Wei
from brownie.exceptions import BrownieEnvironmentError, CompilerError
from brownie.project.build import DEPLOYMENT_KEYS
from brownie.typing import ContractBuildJson, ContractName, Count, PCMap, ProgramCounter
from brownie.utils import bytes_to_hexstring
from brownie.utils.sql import Cursor

from .transaction import TransactionReceipt
from .web3 import _resolve_address, web3

if TYPE_CHECKING:
    from .contract import Contract, ProjectContract

PathMap = dict[str, tuple[HexStr, str]]
Deployment = tuple[ContractBuildJson, dict[str, Any]]

AnyContract = Union["Contract", "ProjectContract"]

_contract_map: Final[dict[ChecksumAddress, AnyContract]] = {}
_revert_refs: Final[list[weakref.ReferenceType]] = []

cur: Final = Cursor(_get_data_folder().joinpath("deployments.db"))
cur.execute("CREATE TABLE IF NOT EXISTS sources (hash PRIMARY KEY, source)")


@final
class TxHistory(metaclass=_Singleton):
    """List-like singleton container that contains TransactionReceipt objects.
    Whenever a transaction is broadcast, the TransactionReceipt is automatically
    added to this container."""

    def __init__(self) -> None:
        self._list: list[TransactionReceipt] = []
        self.gas_profile: Final[dict[str, dict[str, int]]] = {}
        _revert_register(self)

    def __repr__(self) -> str:
        if CONFIG.argv["cli"] == "console":
            return str(self._list)
        return super().__repr__()

    def __getattribute__(self, name: str) -> Any:
        # filter dropped transactions prior to attribute access
        items: list[TransactionReceipt] = super().__getattribute__("_list")
        items = [i for i in items if i.status != -2]
        setattr(self, "_list", items)
        return super().__getattribute__(name)

    def __bool__(self) -> bool:
        return bool(self._list)

    def __contains__(self, item: Any) -> bool:
        return item in self._list

    def __iter__(self) -> Iterator[TransactionReceipt]:
        return iter(self._list)

    def __getitem__(self, key: int) -> TransactionReceipt:
        return self._list[key]

    def __len__(self) -> int:
        return len(self._list)

    def _reset(self) -> None:
        self._list.clear()

    def _revert(self, height: BlockNumber) -> None:
        self._list = [i for i in self._list if i.block_number <= height]  # type: ignore [operator]

    def _add_tx(self, tx: TransactionReceipt) -> None:
        if tx not in self._list:
            self._list.append(tx)

    def clear(self, only_confirmed: bool = False) -> None:
        """
        Clear the list.

        Arguments
        ---------
        only_confirmed : bool, optional
            If True, transactions which are still marked as pending will not be removed.
        """
        if only_confirmed:
            self._list = [i for i in self._list if i.status == -1]
        else:
            self._list.clear()

    def copy(self) -> list[TransactionReceipt]:
        """Returns a shallow copy of the object as a list"""
        return self._list.copy()

    def filter(self, key: Callable | None = None, **kwargs: Any) -> list[TransactionReceipt]:
        """
        Return a filtered list of transactions.

        Arguments
        ---------
        key : Callable, optional
            An optional function to filter with. It should expect one argument and return
            True or False.

        Keyword Arguments
        -----------------
        **kwargs : Any
            Names and expected values for TransactionReceipt attributes.

        Returns
        -------
        List
            A filtered list of TransactionReceipt objects.
        """
        result = (i for i in self._list if all(getattr(i, k) == v for k, v in kwargs.items()))
        return list(result if key is None else filter(key, result))

    def wait(self, key: Callable | None = None, **kwargs: Any) -> None:
        """
        Wait for pending transactions to confirm.

        This method iterates over a list of transactions generated by `TxHistory.filter`,
        waiting until each transaction has confirmed. If no arguments are given, all
        transactions within the container are used.

        Arguments
        ---------
        key : Callable, optional
            An optional function to filter with. It should expect one argument and return
            True or False.

        Keyword Arguments
        -----------------
        **kwargs : Any
            Names and expected values for TransactionReceipt attributes.
        """
        while True:
            pending = next(iter(self.filter(key, status=-1, **kwargs)), None)
            if pending is None:
                return
            pending._confirmed.wait()

    def from_sender(self, account: str) -> list[TransactionReceipt]:
        """Returns a list of transactions where the sender is account"""
        return [i for i in self._list if i.sender == account]

    def to_receiver(self, account: str) -> list[TransactionReceipt]:
        """Returns a list of transactions where the receiver is account"""
        return [i for i in self._list if i.receiver == account]

    def of_address(self, account: str) -> list[TransactionReceipt]:
        """Returns a list of transactions where account is the sender or receiver"""
        return [i for i in self._list if i.receiver == account or i.sender == account]

    def _gas(self, fn_name: str, gas_used: int, is_success: bool) -> None:
        gas = self.gas_profile.setdefault(fn_name, {})
        if not gas:
            gas.update(
                avg=gas_used, high=gas_used, low=gas_used, count=1, count_success=0, avg_success=0
            )
            if is_success:
                gas["count_success"] = 1
                gas["avg_success"] = gas_used
            return
        gas.update(
            avg=(gas["avg"] * gas["count"] + gas_used) // (gas["count"] + 1),
            high=max(gas["high"], gas_used),
            low=min(gas["low"], gas_used),
        )
        gas["count"] += 1
        if is_success:
            count = gas["count_success"]
            gas["count_success"] += 1
            if not gas["avg_success"]:
                gas["avg_success"] = gas_used
            else:
                avg = gas["avg_success"]
                gas["avg_success"] = (avg * count + gas_used) // (count + 1)


@final
class Chain(metaclass=_Singleton):
    """
    List-like singleton used to access block data, and perform actions such as
    snapshotting, mining, and chain rewinds.
    """

    def __init__(self) -> None:
        self._time_offset: int = 0
        self._snapshot_id: int | str | None = None
        self._reset_id: int | str | None = None
        self._current_id: int | str | None = None
        self._undo_lock: Final = threading.Lock()
        self._undo_buffer: Final[list[tuple[int | str, Any, tuple[Any, ...], dict[str, Any]]]] = []
        self._redo_buffer: Final[list[tuple[Any, tuple[Any, ...], dict[str, Any]]]] = []
        self._chainid: int | None = None
        self._block_gas_time: int = -1
        self._block_gas_limit: int = 0

    def __repr__(self) -> str:
        try:
            return f"<Chain object (chainid={self.id}, height={self.height})>"
        except Exception:
            return "<Chain object (disconnected)>"

    def __len__(self) -> int:
        """
        Return the current number of blocks.
        """
        return web3.eth.block_number + 1

    def __getitem__(self, block_number: BlockNumber) -> BlockData | AttributeDict:
        """
        Return information about a block by block number.

        Arguments
        ---------
        block_number : BlockNumber
            Integer of a block number. If the value is negative, the block returned
            is relative to the most recently mined block, e.g. `chain[-1]` returns
            the most recent block.

        Returns
        -------
        BlockData
            web3 block data object
        """
        if not isinstance(block_number, int):
            raise TypeError("Block height must be given as an integer")
        if block_number < 0:
            block_number = web3.eth.block_number + 1 + block_number
        block: BlockData | AttributeDict = web3.eth.get_block(block_number)
        if block["timestamp"] > self._block_gas_time:
            self._block_gas_limit = block["gasLimit"]
            self._block_gas_time = block["timestamp"]
        return block

    def __iter__(self) -> Iterator[BlockData | AttributeDict]:
        get_block = web3.eth.get_block
        for i in range(web3.eth.block_number + 1):
            block: BlockData | AttributeDict = get_block(i)
            yield block

    def new_blocks(
        self, height_buffer: int = 0, poll_interval: int = 5
    ) -> Iterator[BlockData | AttributeDict]:
        """
        Generator for iterating over new blocks.

        Arguments
        ---------
        height_buffer : int, optional
            The number of blocks behind "latest" to return. A higher value means
            more delayed results but less likelihood of uncles.
        poll_interval : int, optional
            Maximum interval between querying for a new block, if the height has
            not changed. Set this lower to detect uncles more frequently.
        """
        if height_buffer < 0:
            raise ValueError("Buffer cannot be negative")

        last_block: BlockData | AttributeDict | None = None
        last_height = 0
        last_poll = 0.0

        get_block = web3.eth.get_block
        while True:
            if last_poll + poll_interval < time.time() or last_height != web3.eth.block_number:
                last_height = web3.eth.block_number
                block: BlockData | AttributeDict = get_block(last_height - height_buffer)
                last_poll = time.time()

                if block != last_block:
                    last_block = block
                    yield last_block
            else:
                time.sleep(1)

    @property
    def height(self) -> BlockNumber:
        return web3.eth.block_number

    @property
    def id(self) -> int:
        if self._chainid is None:
            self._chainid = web3.eth.chain_id
        return self._chainid

    @property
    def block_gas_limit(self) -> Wei:
        if time.time() > self._block_gas_time + 3600:
            block: BlockData | AttributeDict = web3.eth.get_block("latest")
            self._block_gas_limit = block["gasLimit"]
            self._block_gas_time = block["timestamp"]
        return Wei(self._block_gas_limit)

    @property
    def base_fee(self) -> Wei:
        block: BlockData | AttributeDict = web3.eth.get_block("latest")
        return Wei(block["baseFeePerGas"])

    @property
    def priority_fee(self) -> Wei:
        return Wei(web3.eth.max_priority_fee)

    def _revert(self, id_: int | str) -> int | str:
        rpc_client = rpc.Rpc()
        if web3.isConnected() and not web3.eth.block_number and not self._time_offset:
            _notify_registry(0)  # type: ignore [arg-type]
            return rpc_client.snapshot()
        rpc_client.revert(id_)
        id_ = rpc_client.snapshot()
        try:
            self.sleep(0)
        except NotImplementedError:
            pass
        _notify_registry()
        return id_

    def _add_to_undo_buffer(
        self, tx: TransactionReceipt, fn: Any, args: tuple[Any, ...], kwargs: dict[str, Any]
    ) -> None:
        with self._undo_lock:
            tx._confirmed.wait()
            self._undo_buffer.append((self._current_id, fn, args, kwargs))  # type: ignore [arg-type]
            redo_buffer = self._redo_buffer
            if redo_buffer and (fn, args, kwargs) == redo_buffer[-1]:
                redo_buffer.pop()
            else:
                redo_buffer.clear()
            self._current_id = rpc.Rpc().snapshot()
            # ensure the local time offset is correct, in case it was modified by the transaction
            self.sleep(0)

    def _network_connected(self) -> None:
        self._reset_id = None
        try:
            self.reset()
        except NotImplementedError:
            # required for geth dev
            _notify_registry(0)  # type: ignore [arg-type]

    def _network_disconnected(self) -> None:
        self._undo_buffer.clear()
        self._redo_buffer.clear()
        self._snapshot_id = None
        self._reset_id = None
        self._current_id = None
        self._chainid = None
        _notify_registry(0)  # type: ignore [arg-type]

    def get_transaction(self, txid: str | bytes) -> TransactionReceipt:
        """
        Return a TransactionReceipt object for the given transaction hash.
        """
        if not isinstance(txid, str):
            txid = bytes_to_hexstring(txid)
        tx = next((i for i in TxHistory() if i.txid == txid), None)
        return tx or TransactionReceipt(txid, silent=True, required_confs=0)

    def time(self) -> int:
        """Return the current epoch time from the test RPC as an int"""
        return int(time.time() + self._time_offset)

    def sleep(self, seconds: int) -> None:
        """
        Increase the time within the test RPC.

        Arguments
        ---------
        seconds : int
            Number of seconds to increase the time by
        """
        if not isinstance(seconds, int):
            raise TypeError("seconds must be an integer value")
        self._time_offset = int(rpc.Rpc().sleep(seconds))

        if seconds:
            self._redo_buffer.clear()
            self._current_id = rpc.Rpc().snapshot()

    def mine(
        self, blocks: int = 1, timestamp: int | None = None, timedelta: int | None = None
    ) -> BlockNumber:
        """
        Increase the block height within the test RPC.

        Arguments
        ---------
        blocks : int
            Number of new blocks to be mined
        timestamp : int
            Timestamp of the final block being mined. If multiple blocks
            are mined, they will be placed at equal intervals starting
            at `chain.time()` and ending at `timestamp`.
        timedelta : int
            Timedelta for the final block to be mined. If given, the final
            block will have a timestamp of `chain.time() + timedelta`

        Returns
        -------
        BlockNumber
            Current block height
        """
        if not isinstance(blocks, int):
            raise TypeError("`blocks` must be an integer value")

        if timedelta is not None:
            if timestamp is not None:
                raise ValueError("Cannot use both `timestamp` and `timedelta`")

            timestamp = self.time() + timedelta

        if timestamp is None:
            params: list = [[] for _ in range(blocks)]
        elif blocks == 1:
            params = [[timestamp]]
        else:
            now = self.time()
            duration = (timestamp - now) / (blocks - 1)
            params = [[round(now + duration * i)] for i in range(blocks)]

        for i in range(blocks):
            rpc.Rpc().mine(*params[i])

        if timestamp is not None:
            self.sleep(0)

        self._redo_buffer.clear()
        self._current_id = rpc.Rpc().snapshot()
        return web3.eth.block_number

    def snapshot(self) -> None:
        """
        Take a snapshot of the current state of the EVM.

        This action clears the undo buffer.
        """
        self._undo_buffer.clear()
        self._redo_buffer.clear()
        self._snapshot_id = self._current_id = rpc.Rpc().snapshot()

    def revert(self) -> BlockNumber:
        """
        Revert the EVM to the most recently taken snapshot.

        This action clears the undo buffer.

        Returns
        -------
        BlockNumber
            Current block height
        """
        if self._snapshot_id is None:
            raise ValueError("No snapshot set")
        self._undo_buffer.clear()
        self._redo_buffer.clear()
        self._snapshot_id = self._current_id = self._revert(self._snapshot_id)
        return web3.eth.block_number

    def reset(self) -> BlockNumber:
        """
        Revert the EVM to the initial state when loaded.

        This action clears the undo buffer.

        Returns
        -------
        BlockNumber
            Current block height
        """
        self._snapshot_id = None
        self._undo_buffer.clear()
        self._redo_buffer.clear()
        if self._reset_id is None:
            self._reset_id = self._current_id = rpc.Rpc().snapshot()
            _notify_registry(0)  # type: ignore [arg-type]
        else:
            self._reset_id = self._current_id = self._revert(self._reset_id)
        return web3.eth.block_number

    def undo(self, num: int = 1) -> BlockNumber:
        """
        Undo one or more transactions.

        Arguments
        ---------
        num : int, optional
            Number of transactions to undo.

        Returns
        -------
        BlockNumber
            Current block height
        """
        with self._undo_lock:
            if num < 1:
                raise ValueError("num must be greater than zero")
            if not self._undo_buffer:
                raise ValueError("Undo buffer is empty")
            if num > len(self._undo_buffer):
                raise ValueError(f"Undo buffer contains {len(self._undo_buffer)} items")

            for _ in range(num):
                id_, fn, args, kwargs = self._undo_buffer.pop()
                self._redo_buffer.append((fn, args, kwargs))

            self._current_id = self._revert(id_)
            return web3.eth.block_number

    def redo(self, num: int = 1) -> BlockNumber:
        """
        Redo one or more undone transactions.

        Arguments
        ---------
        num : int, optional
            Number of transactions to redo.

        Returns
        -------
        BlockNumber
            Current block height
        """
        with self._undo_lock:
            if num < 1:
                raise ValueError("num must be greater than zero")
            if not self._redo_buffer:
                raise ValueError("Redo buffer is empty")
            if num > len(self._redo_buffer):
                raise ValueError(f"Redo buffer contains {len(self._redo_buffer)} items")

            for _ in range(num):
                fn, args, kwargs = self._redo_buffer.pop()
                fn(*args, **kwargs)

            return web3.eth.block_number


# objects that will update whenever the RPC is reset or reverted must register
# by calling to this function. The must also include _revert and _reset methods
# to receive notifications from this object
def _revert_register(obj: object) -> None:
    _revert_refs.append(weakref.ref(obj))


def _notify_registry(height: BlockNumber | None = None) -> None:
    gc.collect()
    if height is None:
        height = web3.eth.block_number
    for ref in _revert_refs.copy():
        obj = ref()
        if obj is None:
            _revert_refs.remove(ref)
        elif height:
            obj._revert(height)
        else:
            obj._reset()


def _find_contract(address: HexAddress | None) -> AnyContract | None:
    if address is None:
        return None

    address = _resolve_address(address)
    if address in _contract_map:
        return _contract_map[address]
    if "chainid" not in CONFIG.active_network:
        return None

    from brownie.network.contract import Contract

    try:
        return Contract(address)
    except (ValueError, CompilerError):
        return None


def _get_current_dependencies() -> list[ContractName]:
    dependencies = {v._name for v in _contract_map.values()}
    for contract in _contract_map.values():
        dependencies.update(contract._build.get("dependencies", []))
    return sorted(dependencies)


def _add_contract(contract: AnyContract) -> None:
    _contract_map[contract.address] = contract


def _remove_contract(contract: AnyContract) -> None:
    _contract_map.pop(contract.address, None)


def _get_deployment(
    address: HexAddress | None = None,
    alias: ContractName | None = None,
) -> Deployment | tuple[None, None]:
    if address and alias:
        raise ValueError("Passed both params address and alias, should be only one!")
    if address:
        address = _resolve_address(address)
        query = f"address='{address}'"
    elif alias:
        query = f"alias='{alias}'"

    try:
        name = f"chain{CONFIG.active_network['chainid']}"
    except KeyError:
        raise BrownieEnvironmentError("Functionality not available in local environment") from None
    try:
        row = cur.fetchone(f"SELECT * FROM {name} WHERE {query}")
    except OperationalError:
        row = None
    if not row:
        return None, None

    keys = ("address", "alias", "paths") + DEPLOYMENT_KEYS
    build_json: ContractBuildJson = dict(zip(keys, row))  # type: ignore [assignment]
    # when we json.dump the path map, the tuples are encoded as lists so we need to make them tuples again.
    path_map: PathMap = {k: tuple(v) for k, v in build_json.pop("paths", {}).items()}  # type: ignore [typeddict-item]
    sources: dict[str, Any] = {
        i[1]: cur.fetchone("SELECT source FROM sources WHERE hash=?", (i[0],))[0]  # type: ignore [index]
        for i in path_map.values()
    }

    build_json["allSourcePaths"] = {k: path_map[k][1] for k in path_map}
    pc_map: dict[int | str, ProgramCounter] | None = build_json.get("pcMap")  # type: ignore [assignment]
    if isinstance(pc_map, dict):
        build_json["pcMap"] = PCMap({Count(int(k)): pc_map[k] for k in pc_map})

    return build_json, sources


def _add_deployment(
    contract: "Contract",
    alias: ContractName | None = None,
) -> None:
    if "chainid" not in CONFIG.active_network:
        return

    address = _resolve_address(contract.address)
    name = f"chain{CONFIG.active_network['chainid']}"

    cur.execute(
        f"CREATE TABLE IF NOT EXISTS {name} "
        f"(address UNIQUE, alias UNIQUE, paths, {', '.join(DEPLOYMENT_KEYS)})"
    )

    contract_build = contract._build
    if "compiler" not in contract_build:
        # do not replace full contract artifacts with ABI-only ones
        row = cur.fetchone(f"SELECT compiler FROM {name} WHERE address=?", (address,))
        if row and row[0]:
            return

    all_sources = {}
    source_paths: dict = contract_build.get("allSourcePaths", {})
    if source_paths:
        contract_sources: dict = contract._sources
        for key, path in source_paths.items():
            source = contract_sources.get(path)
            if source is None:
                source = Path(path).read_text()
            hash_ = sha1(source.encode()).hexdigest()
            cur.insert("sources", hash_, source)
            all_sources[key] = [hash_, path]

    values = (contract_build.get(i) for i in DEPLOYMENT_KEYS)
    cur.insert(name, address, alias, all_sources, *values)


def _remove_deployment(
    address: HexAddress | None = None,
    alias: ContractName | None = None,
) -> Deployment | tuple[None, None]:
    if address and alias:
        raise ValueError("Passed both params address and alias, should be only one!")
    if address:
        address = _resolve_address(address)
        query = f"address='{address}'"
    elif alias:
        query = f"alias='{alias}'"

    try:
        name = f"chain{CONFIG.active_network['chainid']}"
    except KeyError:
        raise BrownieEnvironmentError("Functionality not available in local environment") from None

    deployment = _get_deployment(address, alias)
    # delete entry from chain{n}
    cur.execute(f"DELETE FROM {name} WHERE {query}")
    # delete all entries from sources matching the contract's source hashes
    if contract := _find_contract(address):
        for key, path in contract._build.get("allSourcePaths", {}).items():
            source = contract._sources.get(path)
            if source is None:
                source = Path(path).read_text()
            hash_ = sha1(source.encode()).hexdigest()
            cur.execute(f"DELETE FROM sources WHERE hash='{hash_}'")

    return deployment



#!/usr/bin/python3

from typing import Any, Final

import faster_eth_utils
from eth_typing import ChecksumAddress

from brownie._c_constants import Decimal, HexBytes

from .datatypes import EthAddress, Fixed, HexString, Wei
from .utils import get_int_bounds

is_hex: Final = faster_eth_utils.is_hex
to_text: Final = faster_eth_utils.to_text

_TEN_DECIMALS: Final = Decimal("1.0000000000")


def to_uint(value: Any, type_str: str = "uint256") -> Wei:
    """Convert a value to an unsigned integer"""
    wei: Wei = Wei(value)
    lower, upper = get_int_bounds(type_str)
    if wei < lower or wei > upper:
        raise OverflowError(f"{value} is outside allowable range for {type_str}")
    return wei


def to_int(value: Any, type_str: str = "int256") -> Wei:
    """Convert a value to a signed integer"""
    wei = Wei(value)
    lower, upper = get_int_bounds(type_str)
    if wei < lower or wei > upper:
        raise OverflowError(f"{value} is outside allowable range for {type_str}")
    return wei


def to_decimal(value: Any) -> Fixed:
    """Convert a value to a fixed point decimal"""
    d: Fixed = Fixed(value)
    if d < -(2**127) or d >= 2**127:
        raise OverflowError(f"{value} is outside allowable range for decimal")
    if d.quantize(_TEN_DECIMALS) != d:
        raise ValueError("Maximum of 10 decimal points allowed")
    return d


def to_address(value: str | bytes) -> ChecksumAddress:
    """Convert a value to an address"""
    return str(EthAddress(value))  # type: ignore [return-value]


def to_bytes(value: Any, type_str: str = "bytes32") -> bytes:
    """Convert a value to bytes"""
    return bytes(HexString(value, type_str))


def to_bool(value: Any) -> bool:
    """Convert a value to a boolean"""
    if not isinstance(value, (int, float, bool, bytes, str)):
        raise TypeError(f"Cannot convert {type(value).__name__} '{value}' to bool")
    if isinstance(value, bytes):
        if not value:
            return False
        value = int(value.hex(), 16)
    if isinstance(value, str) and value.startswith("0x"):
        value = int(value, 16)
    if value not in (0, 1, True, False):
        raise ValueError(f"Cannot convert {type(value).__name__} '{value}' to bool")
    return bool(value)


def to_string(value: Any) -> str:
    """Convert a value to a string"""
    try:
        if isinstance(value, bytes):
            return to_text(hexstr=HexBytes(value).hex())  # type: ignore [arg-type]
        value = str(value)
        if value.startswith("0x") and is_hex(value):
            return to_text(hexstr=value)
        return value
    except UnicodeDecodeError as e:
        raise ValueError(e)


__all__ = ["to_uint", "to_int", "to_decimal", "to_address", "to_bytes", "to_bool", "to_string"]



#!/usr/bin/python3

import decimal
from collections.abc import ItemsView, Iterable, KeysView, Sequence
from typing import Any, Final, Optional, TypeVar, Union, final, overload

try:
    from vyper.exceptions import DecimalOverrideException
except ImportError:
    DecimalOverrideException = BaseException  # regular catch blocks shouldn't catch

import cchecksum
import faster_eth_utils
from eth_typing import ABIComponent, HexStr
from mypy_extensions import mypyc_attr
from typing_extensions import Self

from brownie._c_constants import Decimal, HexBytes, deepcopy, getcontext
from brownie.utils import bytes_to_hexstring

UNITS: Final = {
    "wei": 0,
    "kwei": 3,
    "babbage": 3,
    "mwei": 6,
    "lovelace": 6,
    "gwei": 9,
    "shannon": 9,
    "microether": 12,
    "szabo": 12,
    "milliether": 15,
    "finney": 15,
    "ether": 18,
}

WeiInputTypes = TypeVar("WeiInputTypes", str, float, int, bytes, decimal.Decimal, None)
# This is no longer used within the codebase but we leave it
# in place in case downstream users import it

WeiInputType = str | float | int | bytes | decimal.Decimal | None


to_checksum_address: Final = cchecksum.to_checksum_address

add_0x_prefix: Final = faster_eth_utils.add_0x_prefix
is_hex: Final = faster_eth_utils.is_hex
to_bytes: Final = faster_eth_utils.to_bytes


@final
@mypyc_attr(native_class=False)
class Wei(int):
    """Integer subclass that converts a value to wei and allows comparison against
    similarly formatted values.

    Useful for the following formats:
        * a string specifying the unit: "10 ether", "300 gwei", "0.25 shannon"
        * a large float in scientific notation, where direct conversion to int
          would cause inaccuracy: 8.3e32
        * bytes: b'\xff\xff'
        * hex strings: "0x330124\" """

    def __new__(cls, value: WeiInputType) -> Self:
        return int.__new__(cls, _to_wei(value))

    def __hash__(self) -> int:
        return super().__hash__()

    def __lt__(self, other: Any) -> bool:
        return super().__lt__(_to_wei(other))

    def __le__(self, other: Any) -> bool:
        return super().__le__(_to_wei(other))

    def __eq__(self, other: Any) -> bool:
        try:
            return super().__eq__(_to_wei(other))
        except TypeError:
            return False

    def __ne__(self, other: Any) -> bool:
        try:
            return super().__ne__(_to_wei(other))
        except TypeError:
            return True

    def __ge__(self, other: Any) -> bool:
        return super().__ge__(_to_wei(other))

    def __gt__(self, other: Any) -> bool:
        return super().__gt__(_to_wei(other))

    def __add__(self, other: Any) -> "Wei":
        return Wei(super().__add__(_to_wei(other)))

    def __sub__(self, other: Any) -> "Wei":
        return Wei(super().__sub__(_to_wei(other)))

    def to(self, unit: str) -> "Fixed":
        """
        Returns a converted denomination of the stored wei value.
        Accepts any valid ether unit denomination as string, like:
        "gwei", "milliether", "finney", "ether".

        :param unit: An ether denomination like "ether" or "gwei"
        :return: A 'Fixed' type number in the specified denomination
        """
        try:
            return Fixed(self * Fixed(10) ** -UNITS[unit])
        except KeyError:
            raise TypeError(f'Cannot convert wei to unknown unit: "{unit}". ') from None


def _to_wei(value: WeiInputType) -> int:
    original = value
    if isinstance(value, bytes):
        hexstr = HexBytes(value).hex().removeprefix("0x")
        return int(hexstr, 16) if hexstr else 0
    if not value or value == "0x":
        return 0
    if isinstance(value, float) and "e+" in (string := str(value)):
        num_str, dec = string.split("e+")
        num = num_str.split(".") if "." in num_str else [num_str, ""]
        return int(num[0] + num[1][: int(dec)] + "0" * (int(dec) - len(num[1])))
    if not isinstance(value, str):
        return _return_int(original, value)
    elif value.startswith("0x"):
        return int(value, 16)
    for unit, decimals in UNITS.items():
        if f" {unit}" not in value:
            continue
        num_str = value.split(" ")[0]
        num = num_str.split(".") if "." in num_str else [num_str, ""]
        return int(num[0] + num[1][:decimals] + "0" * (decimals - len(num[1])))
    return _return_int(original, value)


def _return_int(original: Any, value: Any) -> int:
    try:
        return int(value)
    except ValueError:
        raise TypeError(f"Cannot convert {type(original).__name__} '{original}' to wei.")


@final
@mypyc_attr(native_class=False)
class Fixed(decimal.Decimal):
    """
    Decimal subclass that allows comparison against strings, integers and Wei.

    Raises TypeError when operations are attempted against floats.
    """

    def __new__(cls, value: Any) -> Self:
        return Decimal.__new__(cls, _to_fixed(value))

    def __repr__(self) -> str:
        return f"Fixed('{str(self)}')"

    def __hash__(self) -> int:
        return super().__hash__()

    def __lt__(self, other: Any) -> bool:
        return super().__lt__(_to_fixed(other))

    def __le__(self, other: Any) -> bool:
        return super().__le__(_to_fixed(other))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, float):
            raise TypeError("Cannot compare to floating point - use a string instead")
        try:
            return super().__eq__(_to_fixed(other))
        except TypeError:
            return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, float):
            raise TypeError("Cannot compare to floating point - use a string instead")
        try:
            return super().__ne__(_to_fixed(other))
        except TypeError:
            return True

    def __ge__(self, other: Any) -> bool:
        return super().__ge__(_to_fixed(other))

    def __gt__(self, other: Any) -> bool:
        return super().__gt__(_to_fixed(other))

    def __add__(self, other: Any) -> "Fixed":
        return Fixed(super().__add__(_to_fixed(other)))

    def __sub__(self, other: Any) -> "Fixed":
        return Fixed(super().__sub__(_to_fixed(other)))


def _to_fixed(value: Any) -> decimal.Decimal:
    if isinstance(value, float):
        raise TypeError("Cannot convert float to decimal - use a string instead")

    if isinstance(value, (str, bytes)):
        try:
            value = Wei(value)
        except TypeError:
            pass
    try:
        try:
            # until vyper v0.3.1 we can mess with the precision
            ctx = getcontext()
            ctx.prec = 78
        except DecimalOverrideException:
            pass  # vyper set the precision, do nothing.
        return Decimal(value)
    except Exception as e:
        raise TypeError(f"Cannot convert {type(value).__name__} '{value}' to decimal.") from e


@final
@mypyc_attr(native_class=False)
class EthAddress(str):
    """String subclass that raises TypeError when compared to a non-address."""

    def __new__(cls, value: Any) -> Self:
        converted_value: HexStr
        if isinstance(value, str):
            converted_value = value  # type: ignore [assignment]
        elif isinstance(value, bytes):
            converted_value = bytes_to_hexstring(value)
        else:
            converted_value = str(value)  # type: ignore [assignment]
        converted_value = add_0x_prefix(converted_value)
        try:
            converted_value = to_checksum_address(converted_value)
        except ValueError:
            raise ValueError(f"{value!r} is not a valid ETH address") from None
        return str.__new__(cls, converted_value)

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, other: Any) -> bool:
        return _address_compare(self, other)

    def __ne__(self, other: Any) -> bool:
        return not _address_compare(self, other)


def _address_compare(a: str, b: Any) -> bool:
    bstr = str(b)
    if not bstr.startswith("0x") or not is_hex(bstr) or len(bstr) != 42:
        raise TypeError(f"Invalid type for comparison: '{bstr}' is not a valid address")
    return a.lower() == bstr.lower()


@final
@mypyc_attr(native_class=False)
class HexString(bytes):
    """Bytes subclass for hexstring comparisons. Raises TypeError if compared to
    a non-hexstring. Evaluates True for hexstrings with the same value but differing
    leading zeros or capitalization."""

    def __new__(cls, value: Any, type_str: str) -> Self:
        return bytes.__new__(cls, _to_bytes(value, type_str))

    def __eq__(self, other: Any) -> bool:
        return _hex_compare(self.hex(), other)

    def __ne__(self, other: Any) -> bool:
        return not _hex_compare(self.hex(), other)

    def __str__(self) -> HexStr:
        return f"0x{self.hex()}"  # type: ignore [return-value]

    def __repr__(self) -> HexStr:
        return str(self)  # type: ignore [return-value]


def _hex_compare(a: str, b: Any) -> bool:
    bstr = str(b)
    if not bstr.startswith("0x") or not is_hex(bstr):
        raise TypeError(f"Invalid type for comparison: '{bstr}' is not a valid hex string")
    return a.lstrip("0x").lower() == bstr.lstrip("0x").lower()


def _to_bytes(value: Any, type_str: str = "bytes32") -> bytes:
    """Convert a value to bytes"""
    if isinstance(value, bool) or not isinstance(value, (bytes, str, int)):
        raise TypeError(f"Cannot convert {type(value).__name__} '{value}' to {type_str}")
    value = _to_hex(value)
    if type_str == "bytes":
        return to_bytes(hexstr=value)
    if type_str == "byte":
        type_str = "bytes1"
    size = int(type_str.strip("bytes"))
    if size < 1 or size > 32:
        raise ValueError(f"Invalid type: {type_str}")
    try:
        return int(value, 16).to_bytes(size, "big")
    except OverflowError:
        raise OverflowError(f"'{value}' exceeds maximum length for {type_str}")


def _to_hex(value: Any) -> HexStr:
    """Convert a value to a hexstring"""
    if isinstance(value, bytes):
        return bytes_to_hexstring(value)
    if isinstance(value, int):
        return HexStr(hex(value))
    if isinstance(value, str):
        if value in ("", "0x"):
            return HexStr("0x00")
        if is_hex(value):
            return add_0x_prefix(value)
    raise ValueError(f"Cannot convert {type(value).__name__} '{value}' to a hex string")


@final
@mypyc_attr(native_class=False)
class ReturnValue(tuple):
    """Tuple subclass with dict-like functionality, used for iterable return values."""

    _abi: list[ABIComponent] | None = None
    _dict: dict[str, Any] = {}

    def __new__(
        cls,
        values: Iterable[Any],
        abi: Sequence[ABIComponent] | None = None,
    ) -> "ReturnValue":
        values = list(values)
        for i, value in enumerate(values):
            if isinstance(value, (tuple, list)) and not isinstance(value, ReturnValue):
                if abi is not None and "components" in (value_abi := abi[i]):
                    if value_abi["type"] == "tuple":
                        # tuple
                        values[i] = ReturnValue(value, value_abi["components"])
                    else:
                        # array of tuples
                        inner_abi = value_abi.copy()
                        length = len(value)
                        inner_abi["type"] = inner_abi["type"].rsplit("[", maxsplit=1)[0]
                        final_abi = [deepcopy(inner_abi) for i in range(length)]
                        if inner_abi.get("name"):
                            name = inner_abi["name"]
                            for x in range(length):
                                final_abi[x]["name"] = f"{name}[{x}]"

                        values[i] = ReturnValue(value, final_abi)
                else:
                    # array
                    values[i] = ReturnValue(value)

        self = tuple.__new__(cls, values)
        self._abi = list(abi) if abi else []
        self._dict = {i.get("name", "") or f"arg[{c}]": values[c] for c, i in enumerate(self._abi)}

        return self

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, other: Any) -> bool:
        return _kwargtuple_compare(self, other)

    def __ne__(self, other: Any) -> bool:
        return not _kwargtuple_compare(self, other)

    @overload  # type: ignore [override]
    def __getitem__(self, key: int) -> Any: ...
    @overload
    def __getitem__(self, key: str) -> Any: ...
    @overload
    def __getitem__(
        self,
        key: "slice[Optional[int], Optional[int], Optional[int]]",
    ) -> "ReturnValue": ...
    def __getitem__(
        self,
        key: Union[str, int, "slice[Optional[int], Optional[int], Optional[int]]"],
    ) -> Any:
        if type(key) is slice:
            abi = self._abi
            result = super().__getitem__(key)
            if abi is None:
                return ReturnValue(result)
            item_abi = deepcopy(abi)[key]
            return ReturnValue(result, item_abi)
        if isinstance(key, int):
            return super().__getitem__(key)
        return self._dict[key]

    def __contains__(self, value: Any) -> bool:
        return self.count(value) > 0

    def count(self, value: Any) -> int:
        """ReturnValue.count(value) -> integer -- return number of occurrences of value"""
        count = 0
        for item in self:
            try:
                if _kwargtuple_compare(item, value):
                    count += 1
            except TypeError:
                continue
        return count

    def dict(self) -> dict[str, Any]:
        """ReturnValue.dict() -> a dictionary of ReturnValue's named items"""
        response = {}
        for k, v in self._dict.items():
            if isinstance(v, ReturnValue) and v._abi:
                response[k] = v.dict()
            else:
                response[k] = v
        return response

    def index(self, value: Any, start: int = 0, stop: Any = None) -> int:  # type: ignore [override]
        """ReturnValue.index(value, [start, [stop]]) -> integer -- return first index of value.
        Raises ValueError if the value is not present."""
        if stop is None:
            stop = len(self)
        for i in range(start, stop):
            try:
                if _kwargtuple_compare(self[i], value):
                    return i
            except TypeError:
                continue
        raise ValueError(f"{value} is not in ReturnValue")

    def items(self) -> ItemsView[str, Any]:
        """ReturnValue.items() -> a set-like object providing a view on ReturnValue's named items"""
        return self._dict.items()

    def keys(self) -> KeysView[str]:
        """ReturnValue.keys() -> a set-like object providing a view on ReturnValue's keys"""
        return self._dict.keys()


def _kwargtuple_compare(a: Any, b: Any) -> bool:
    if not isinstance(a, (tuple, list, ReturnValue)):
        types_ = {type(a), type(b)}
        if types_.intersection((bool, type(None))):
            return a is b
        if types_.intersection((dict, EthAddress, HexString)):
            return a == b
        return _convert_str(a) == _convert_str(b)
    if not isinstance(b, (tuple, list, ReturnValue)) or len(b) != len(a):
        return False
    return all(_kwargtuple_compare(ai, bi) for ai, bi in zip(a, b))


def _convert_str(value: Any) -> Wei:
    if not isinstance(value, str):
        if not hasattr(value, "address"):
            return value
        value = value.address
    if value.startswith("0x"):
        return "0x" + value.lstrip("0x").lower()
    if value.count(" ") != 1:
        return value
    try:
        return Wei(value)
    except (ValueError, TypeError):
        return value



from collections.abc import Callable, Sequence
from typing import Any, final

from web3 import Web3
from web3.exceptions import ExtraDataLengthError
from web3.middleware import geth_poa_middleware
from web3.types import RPCEndpoint

from brownie.network.middlewares import BrownieMiddlewareABC


@final
class GethPOAMiddleware(BrownieMiddlewareABC):
    @classmethod
    def get_layer(cls, w3: Web3, network_type: str) -> int | None:
        # NOTE: We check `0` because Ganache sometimes injects a block of their own.
        # It doesn't have the extra data that we are checking for.
        # We also check `"latest"` because On Polygon networks, anvil in forked
        # development mode doesn't throw ExtraDataLengthError on the first block.
        block_idents = ("latest",) if network_type == "live" else (0, "latest")
        try:
            for block_ident in block_idents:
                w3.eth.get_block(block_ident)
            return None
        except ExtraDataLengthError:
            return -1

    def process_request(
        self,
        make_request: Callable,
        method: RPCEndpoint,
        params: Sequence[Any],
    ) -> dict[str, Any]:
        middleware_fn = geth_poa_middleware(make_request, self.w3)
        return middleware_fn(method, params)



#!/usr/bin/python3
# mypy: disable-error-code="union-attr"

import time
import warnings
from collections import OrderedDict
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence, ValuesView
from pathlib import Path
from threading import Lock, Thread
from typing import TYPE_CHECKING, Any, Final, Generic, TypeVar, Union, cast, final, overload

import eth_event
from eth_event import EventError
from eth_event.main import DecodedEvent, NonDecodedEvent, TopicMapData, _TraceStep
from eth_typing import ABIElement, AnyAddress, ChecksumAddress, HexStr
from ujson import JSONDecodeError
from web3._utils import filters
from web3.datastructures import AttributeDict

from brownie._c_constants import ujson_dump, ujson_load
from brownie._config import _get_data_folder
from brownie._singleton import _Singleton
from brownie.convert.datatypes import ReturnValue
from brownie.convert.normalize import format_event
from brownie.exceptions import EventLookupError
from brownie.typing import FormattedEvent, Selector
from brownie.utils import hexbytes_to_hexstring

from .web3 import ContractEvent, web3

if TYPE_CHECKING:
    from .contract import Contract


TopicMap = dict[HexStr, TopicMapData]
DeploymentTopics = dict[ChecksumAddress, TopicMap]

EventData = OrderedDict[str, Any]
"""An OrderedDict which contains the indexed args for a single on-chain event."""


@final
class EventDict:
    """
    Dict/list hybrid container, base class for all events fired in a transaction.
    """

    def __init__(self, events: Iterable[FormattedEvent] | None = None) -> None:
        """Instantiates the class.

        Args:
            events: event data as supplied by eth_event.decode_logs or eth_event.decode_trace"""
        events = list(events) if events else []

        ordered: list[Event] = [
            _EventItem(
                i["name"],
                i["address"],
                [OrderedDict((x["name"], x["value"]) for x in i["data"])],
                (pos,),
            )
            for pos, i in enumerate(events)
        ]
        self._ordered: Final = ordered

        _dict: OrderedDict[str, Events]
        _dict = OrderedDict()
        for event in ordered:
            event_name = event.name
            if event_name not in _dict:
                events_for_name = [i for i in ordered if i.name == event_name]
                _dict[event_name] = _EventItem(
                    event_name,
                    None,
                    events_for_name,
                    tuple(i.pos[0] for i in events_for_name),
                )

        self._dict: Final = _dict

    def __repr__(self) -> str:
        return str(self)

    def __bool__(self) -> bool:
        return bool(self._ordered)

    def __contains__(self, name: str) -> bool:
        """returns True if an event fired with the given name."""
        return name in self._dict

    @overload
    def __getitem__(self, key: int) -> "Event":
        """returns the n'th event that was fired"""

    @overload
    def __getitem__(self, key: str) -> "Events":
        """returns a _EventItem dict of all events where name == key"""

    def __getitem__(self, key: str | int) -> Union["Event", "Events"]:
        """if key is int: returns the n'th event that was fired
        if key is str: returns a _EventItem dict of all events where name == key"""
        if isinstance(key, int):
            try:
                return self._ordered[key]
            except IndexError:
                raise EventLookupError(
                    f"Index out of range - only {len(self._ordered)} events fired"
                ) from None
        elif isinstance(key, str):
            try:
                return self._dict[key]
            except KeyError:
                raise EventLookupError(f"Event '{key}' did not fire.") from None
        else:
            raise TypeError(f"Invalid key type '{type(key)}' - can only use strings or integers")

    def __iter__(self) -> Iterator["Event"]:
        return iter(self._ordered)

    def __len__(self) -> int:
        """returns the number of events that fired."""
        return len(self._ordered)

    def __str__(self) -> str:
        return str({k: [i[0] for i in v._ordered] for k, v in self._dict.items()})

    def count(self, name: str) -> int:
        """EventDict.count(name) -> integer -- return number of occurrences of name"""
        return len(self._dict.get(name, ()))

    def items(self) -> list[tuple[str, "Events"]]:
        """EventDict.items() -> a list object providing a view on EventDict's items"""
        return list(self._dict.items())

    def keys(self) -> list[str]:
        """EventDict.keys() -> a list object providing a view on EventDict's keys"""
        return list(self._dict.keys())

    def values(self) -> ValuesView["Events"]:
        """EventDict.values() -> a list object providing a view on EventDict's values"""
        return self._dict.values()


_TData = TypeVar("_TData", "Event", EventData)


@final
class _EventItem(Generic[_TData]):
    """
    Dict/list hybrid container, represents one or more events with the same name
    that were fired in a transaction.

    Attributes
    ----------
    name : str
        Name of the event.
    address : ChecksumAddress
        Address where this event fired. When the object represents more than one event,
        this value is set to `None`.
    pos : Tuple[int, ...]
        Tuple of indexes where this event fired.
    """

    def __init__(
        self,
        name: str,
        address: ChecksumAddress | None,
        event_data: list[_TData],
        pos: tuple[int, ...],
    ) -> None:
        self.name: Final = name
        self.address: Final = address
        self._ordered: Final = event_data
        self.pos: Final = pos

    @overload
    def __getitem__(self, key: int) -> _TData:
        """returns the n'th event that was fired with this name"""

    @overload
    def __getitem__(self, key: str) -> Any:
        """returns the value of data field 'key' from the 1st event within the container"""

    def __getitem__(self, key: int | str) -> _TData | Any:
        """if key is int: returns the n'th event that was fired with this name
        if key is str: returns the value of data field 'key' from the 1st event
        within the container"""
        if isinstance(key, int):
            try:
                return self._ordered[key]
            except IndexError:
                raise EventLookupError(
                    f"Index out of range - only {len(self._ordered)} '{self.name}' events fired"
                )
        elif isinstance(key, str):
            first = self._ordered[0]
            if key in first:
                return first[key]
            if f"{key} (indexed)" in first:
                return first[f"{key} (indexed)"]
            valid_keys = ", ".join(self.keys())
            raise EventLookupError(
                f"Unknown key '{key}' - the '{self.name}' event includes these keys: {valid_keys}"
            )
        else:
            raise TypeError(f"Invalid key type '{type(key)}' - can only use strings or integers")

    def __contains__(self, name: str) -> bool:
        """returns True if this event contains a value with the given name."""
        return name in self._ordered[0]

    def __len__(self) -> int:
        """returns the number of events held in this container."""
        return len(self._ordered)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        data = self._ordered
        if len(data) == 1:
            return str(data[0])
        return str([i[0] for i in data])  # type: ignore [index]

    def __iter__(self) -> Iterator[_TData]:
        return iter(self._ordered)

    def __eq__(self, other: object) -> bool:
        if len(self._ordered) == 1:
            if isinstance(other, (tuple, list, ReturnValue)):
                # sequences compare directly against the event values
                return self._ordered[0].values() == other
            return other == self._ordered[0]
        return other == self._ordered

    def items(self) -> ReturnValue:
        """_EventItem.items() -> a list object providing a view on _EventItem[0]'s items"""
        return ReturnValue((i, self[i]) for i in self.keys())

    def keys(self) -> ReturnValue:
        """_EventItem.keys() -> a list object providing a view on _EventItem[0]'s keys"""
        return ReturnValue(i.replace(" (indexed)", "") for i in self._ordered[0].keys())

    def values(self) -> ReturnValue:
        """_EventItem.values() -> a list object providing a view on _EventItem[0]'s values"""
        return ReturnValue(self._ordered[0].values())  # type: ignore [arg-type]


Event = _EventItem[EventData]
"""An _EventItem which represents a single event."""

Events = _EventItem[Event]
"""An _EventItem which represents a collection of events which share the same event name."""


@final
class _EventWatchData:
    """
    Class containing the data needed to check, time, and execute callbacks on a specified event.
    """

    def __init__(
        self,
        event: ContractEvent,
        callback: Callable[[AttributeDict], None],
        delay: float = 2.0,
        repeat: bool = True,
    ) -> None:
        # Args
        self.event: Final[ContractEvent] = event
        self._callbacks_list: list[dict] = []
        self.delay: float = delay
        # Members
        self._event_filter: Final[filters.LogFilter] = event.create_filter(
            fromBlock=(web3.eth.block_number - 1)
        )
        self._cooldown_time_over: bool = False
        self.timer = time.time()
        self.add_callback(callback, repeat)

    def get_new_events(self) -> list["filters.LogReceipt"]:
        """
        Retrieves and return the events that occurred between now and the last function call.

        Returns:
            [List[LogReceipt]]: List of the retrieved events
        """
        return self._event_filter.get_new_entries()

    def reset_timer(self) -> None:
        """Resets the 'self.timer' member variable to the current time."""
        self.timer = time.time()

    def add_callback(self, callback: Callable[[AttributeDict], None], repeat: bool = True) -> None:
        """Adds a new callback instruction to execute on new events"""
        if not callable(callback):
            raise TypeError("'callback' argument MUST be a callable object.")
        self._callbacks_list.append({"function": callback, "repeat": repeat})

    def update_delay(self, new_delay: float) -> None:
        """
        Changes the 'delay' member variable if the 'new_delay'
        parameter is inferior to the current one
        """
        self.delay = min(self.delay, new_delay)

    def _trigger_callbacks(self, events_data: list[filters.LogReceipt]) -> list[Thread]:
        """
        Given a list of event as a parameter, creates a thread for each callback
        present in 'self._callbacks_list', stores the thread in a list which is returned.
        Removes non-repeating callbacks from the callback list

        Args:
            events_data (List[AttributeDict]): The list of event to iterate on.
        """

        def _map_callback_on_list(
            callback: Callable, data_to_map: list[filters.LogReceipt]
        ) -> None:
            list(map(callback, data_to_map))

        self.cooldown_time_over = False
        threads: list[Thread] = []
        for callback in self._callbacks_list:
            # Creates a thread for each callback
            threads.append(
                Thread(
                    target=_map_callback_on_list,
                    args=(
                        callback["function"],
                        events_data,
                    ),
                    daemon=True,
                )
            )
            threads[-1].start()
        # Remove non-repeating callbacks from list
        self._callbacks_list = [cb for cb in self._callbacks_list if cb.get("repeat")]
        return threads

    @property
    def time_left(self) -> float:
        """Computes and returns the difference between the self.delay variable
        and the time between now and the last callback_trigger_time.

        Returns:
            float: Time difference between self.delay and the time between
            now and the last callback_trigger_time.
        """
        return max(0.0, self.delay - (time.time() - self.timer))


class EventWatcher(metaclass=_Singleton):
    """
    Singleton class containing methods to set callbacks on user-specified events.
    This class is multi-threaded :
        - The main thread (original process) activates a sub-thread and can be used
        to add callback instructions on specific events.
        - The sub-thread looks for new events among the ones with a callback set.
        When found, calls a method that creates new threads to run the callback
        instructions with the event(s) data as a parameter.
    """

    def __init__(self) -> None:
        self.target_list_lock: Lock = Lock()
        self.target_events_watch_data: dict[str, _EventWatchData] = {}
        self._kill: bool = False
        self._has_started: bool = False
        self._watcher_thread = Thread(target=self._loop, daemon=True)

    def __del__(self) -> None:
        self.stop()

    def stop(self, wait: bool = True) -> None:
        """
        Stops the running threads. This function does not reset the instance
        to its initial state. If that is your goal, check the 'reset' method.

        Args:
            wait (bool, optional): Whether to wait for thread to join within the function.
                Defaults to True.
        """
        self._kill = True
        if wait is True and self._watcher_thread.is_alive():
            self._watcher_thread.join()
        self._has_started = False

    def reset(self) -> None:
        """Stops the running threads and reset the instance to its initial state"""
        self.stop()
        self._setup()

    def add_event_callback(
        self,
        event: ContractEvent,
        callback: Callable[[AttributeDict], None],
        delay: float = 2.0,
        repeat: bool = True,
    ) -> None:
        """
        Adds a callback instruction for the specified event.

        Args:
            event (ContractEvent): The ContractEvent instance to watch for.
            callback (Callable[[AttributeDict], None]): The function to be called
                when a new 'event' is detected.
            delay (float, optional): The delay between each check for new 'event'(s).
                Defaults to 2.0.
            repeat (bool, optional): Whether to repeat the callback or not (if False,
                the callback will be called once only). Defaults to True.

        Raises:
            TypeError: Raises when the parameter 'callback' is not a callable object.
        """
        if not callable(callback):
            raise TypeError("Argument 'callback' argument must be a callable.")
        delay = max(delay, 0.05)
        self.target_list_lock.acquire()  # lock
        # Key referring to this specific event (event.address is the address
        # of the contract to which the event is linked)
        event_watch_data_key = f"{str(event.address)}+{event.event_name}"
        if self.target_events_watch_data.get(event_watch_data_key) is None:
            # If the _EventWatchData for 'event' does not exist, creates it.
            self.target_events_watch_data[event_watch_data_key] = _EventWatchData(
                event, callback, delay, repeat
            )
        else:
            # Adds a new callback to the already existing _EventWatchData.
            self.target_events_watch_data[event_watch_data_key].add_callback(callback, repeat)
            if repeat is True:
                # Updates the delay between each check calling the
                # _EventWatchData.update_delay function
                self.target_events_watch_data[event_watch_data_key].update_delay(delay)
        self.target_list_lock.release()  # unlock
        # Start watch if not done
        if self._has_started is False:
            self._start_watch()

    def _setup(self) -> None:
        """Sets up the EventWatcher instance member variables so it is ready to run"""
        self.target_list_lock.acquire()
        self.target_events_watch_data.clear()
        self.target_list_lock.release()
        self._kill = False
        self._has_started = False
        self._watcher_thread = Thread(target=self._loop, daemon=True)

    def _start_watch(self) -> None:
        """Starts the thread running the _loop function"""
        self._watcher_thread.start()
        self._has_started = True

    def _loop(self) -> None:
        """
        Watches for new events. Whenever new events are detected, calls the
        '_EventWatchData._trigger_callbacks' function to run the callbacks instructions
        (in separate threads) on the detected events data.
        """
        workers_list: list[Thread] = []

        while not self._kill:
            try:
                sleep_time: float = 1.0  # Max sleep time.
                self.target_list_lock.acquire()  # lock
                for _, elem in self.target_events_watch_data.items():
                    # If cooldown is not over :
                    #   skip and store time left before next check if needed.
                    time_left = elem.time_left
                    if time_left > 0:
                        sleep_time = min(sleep_time, time_left)
                        continue
                    # Check for new events & execute callback async if some are found
                    latest_events = elem.get_new_events()
                    if len(latest_events) != 0:
                        workers_list += elem._trigger_callbacks(latest_events)
                    elem.reset_timer()
                    # after elem.reset_timer elem.time_left is approximately elem.delay
                    sleep_time = min(sleep_time, elem.time_left)
            finally:
                self.target_list_lock.release()  # unlock
                # Remove dead threads from the workers_list
                workers_list = list(filter(lambda x: x.is_alive(), workers_list))
                time.sleep(sleep_time)

        # Join running threads when leaving function.
        for worker_instance in workers_list:
            worker_instance.join(timeout=30)
            if worker_instance.is_alive():
                worker_name = worker_instance.getName()
                warnings.warn(
                    message=f"Callback execution ({worker_name}) could not be joined.",
                    category=RuntimeWarning,
                )


def __get_path() -> Path:
    return _get_data_folder().joinpath("topics.json")


def _get_topics(abi: list[ABIElement]) -> dict[str, HexStr]:
    topic_map = eth_event.get_topic_map(abi)

    updated_topics = _topics.copy()

    for key, value in topic_map.items():
        if key not in updated_topics:
            # new event topic
            updated_topics[key] = value
        elif value == updated_topics[key]:
            # existing event topic, nothing has changed
            continue
        elif not any(i["indexed"] for i in updated_topics[key]["inputs"]):
            # existing topic, but the old abi has no indexed events - keep the new one
            updated_topics[key] = value

    if updated_topics != _topics:
        _topics.update(updated_topics)
        with __get_path().open("w") as fp:
            ujson_dump(updated_topics, fp, sort_keys=True, indent=2)

    return {v["name"]: k for k, v in topic_map.items()}


def _add_deployment_topics(address: ChecksumAddress, abi: list[ABIElement]) -> None:
    _deployment_topics[address] = eth_event.get_topic_map(abi)


def _decode_logs(
    # Mapping is included so we don't break dependent lib ypricemagic with this change
    logs: list[_EventItem] | list[Mapping[str, Any]],
    contracts: dict[ChecksumAddress, "Contract"] | None = None,
) -> EventDict:
    if not logs:
        return EventDict()

    idx = 0
    events: list[DecodedEvent | NonDecodedEvent] = []
    while True:
        address: ChecksumAddress = logs[idx]["address"]
        try:
            new_idx = logs.index(
                next(i for i in logs[idx:] if i["address"] != address)  # type: ignore [misc]
            )
            log_slice = logs[idx:new_idx]
            idx = new_idx
        except StopIteration:
            log_slice = logs[idx:]

        topics_map = _deployment_topics.get(address, _topics)
        for item in log_slice:
            if contracts:
                contract = contracts[address]
                if contract is not None:
                    note = _decode_ds_note(item, contract)
                    if note is not None:
                        events.append(note)
                        continue
            try:
                events.extend(
                    eth_event.decode_logs([item], topics_map, allow_undecoded=True)  # type: ignore [call-overload]
                )
            except EventError as exc:
                warnings.warn(f"{address}: {exc}")

        if log_slice[-1] == logs[-1]:
            break

    return EventDict(format_event(event) for event in events)


def _decode_ds_note(
    log: _EventItem | Mapping[str, Any], contract: "Contract"
) -> DecodedEvent | None:
    # ds-note encodes function selector as the first topic
    # TODO double check typing for `log` input
    topic0 = log.topics[0]
    selector, tail = topic0[:4], topic0[4:]
    selector_hexstr = Selector(hexbytes_to_hexstring(selector))
    if selector_hexstr not in contract.selectors or sum(tail):
        return None
    name = contract.selectors[selector_hexstr]
    log_data = log.data
    data = bytes.fromhex(log_data[2:]) if isinstance(log_data, str) else log_data
    # data uses ABI encoding of [uint256, bytes] or [bytes] in different versions
    # instead of trying them all, assume the payload starts from selector
    try:
        func, args = contract.decode_input(data[data.index(selector) :])
    except ValueError:
        return None
    selector_hexstr = Selector(hexbytes_to_hexstring(selector))
    inputs = contract.get_method_object(selector_hexstr).abi["inputs"]
    return DecodedEvent(
        name=name,
        address=cast(ChecksumAddress, log.address),
        decoded=True,
        data=[
            {"name": abi["name"], "type": abi["type"], "value": arg, "decoded": True}
            for arg, abi in zip(args, inputs)
        ],
    )


def _decode_trace(trace: Sequence[_TraceStep], initial_address: AnyAddress) -> EventDict:
    if not trace:
        return EventDict()

    events = eth_event.decode_traceTransaction(
        trace if type(trace) is list else list(trace),
        _topics,
        allow_undecoded=True,
        initial_address=initial_address,
    )
    return EventDict(format_event(event) for event in events)


# dictionary of event topic ABIs specific to a single contract deployment
_deployment_topics: Final[DeploymentTopics] = {}

# EventWatcher program instance
event_watcher: Final = EventWatcher()

try:
    with __get_path().open() as fp:
        __topics = ujson_load(fp)
except (FileNotFoundError, JSONDecodeError):
    __topics = None

# general event topic ABIs for decoding events on unknown contracts
_topics: Final[TopicMap] = __topics or {}
del __topics



from collections.abc import Callable, Sequence
from typing import Any, final

from web3 import Web3
from web3.types import RPCEndpoint

from brownie.network.middlewares import BrownieMiddlewareABC


@final
class TxRevertCatcherMiddleware(BrownieMiddlewareABC):
    """
    Middleware to handle reverting transactions, bypasses web3 error formatting.

    As of web3.py version 5.13.0, a new error formatting middleware was added by default
    `raise_solidity_error_on_revert` which when a `eth_call` or `eth_estimateGas` tx
    raises a `ContractLogicError` instead of providing us with an RPCError dictionary.
    """

    @classmethod
    def get_layer(cls, w3: Web3, network_type: str) -> int | None:
        return -1

    def process_request(
        self,
        make_request: Callable,
        method: RPCEndpoint,
        params: Sequence[Any],
    ) -> dict[str, Any]:
        """Raise a ValueError when RPC.eth_call or RPC.eth_estimateGas errors."""
        result = make_request(method, params)
        if method in {"eth_call", "eth_estimateGas"} and "error" in result:
            raise ValueError(result["error"])
        return result



#!/usr/bin/python3
# mypy: disable-error-code="index,typeddict-unknown-key"

from typing import Any, Dict, Final, List, Optional, Union, cast

import solcast
from eth_typing import ABIElement, HexStr

from brownie._c_constants import Path, Version, deepcopy, sha1, ujson_loads
from brownie._config import _get_data_folder
from brownie.exceptions import UnsupportedLanguage
from brownie.project import sources
from brownie.project.compiler.solidity import (  # NOQA: F401
    find_best_solc_version,
    find_solc_versions,
    install_solc,
    set_solc_version,
)
from brownie.project.compiler.utils import _get_alias, merge_natspec
from brownie.project.compiler.vyper import find_vyper_versions, set_vyper_version
from brownie.typing import (
    CompilerConfig,
    ContractBuildJson,
    ContractName,
    EvmVersion,
    InputJson,
    InputJsonSolc,
    InputJsonVyper,
    InterfaceBuildJson,
    InterfaceSource,
    Language,
    OptimizerSettings,
    SettingsSolc,
    SourcesDict,
)
from brownie.utils import notify

from . import solidity, vyper

STANDARD_JSON: Final[InputJson] = {  # type: ignore [assignment]
    "language": None,
    "sources": {},
    "settings": {
        "outputSelection": {
            "*": {
                "*": ["abi", "devdoc", "evm.bytecode", "evm.deployedBytecode", "userdoc"],
                "": ["ast"],
            }
        },
        "evmVersion": None,
        "remappings": [],
    },
}

EvmVersionSpec = Union[EvmVersion, dict[Language, EvmVersion | None]]


# C constants
_from_standard_output: Final = solcast.from_standard_output


def compile_and_format(
    contract_sources: dict[str, str],
    solc_version: str | None = None,
    vyper_version: str | None = None,
    optimize: bool | None = True,
    runs: int | None = 200,
    evm_version: EvmVersionSpec | None = None,
    silent: bool = True,
    allow_paths: str | None = None,
    interface_sources: dict[str, str] | None = None,
    remappings: list[str] | str | None = None,
    optimizer: OptimizerSettings | None = None,
    viaIR: bool | None = None,
) -> dict[ContractName, ContractBuildJson]:
    """Compiles contracts and returns build data.

    Args:
        contract_sources: a dictionary in the form of {'path': "source code"}
        solc_version: solc version to compile with (use None to set via pragmas)
        optimize: (deprecated) enable solc optimizer
        runs: (deprecated) optimizer runs
        evm_version: evm version to compile for
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path
        interface_sources: dictionary of interfaces as {'path': "source code"}
        remappings: list of solidity path remappings
        optimizer: dictionary of solidity optimizer settings
        viaIR: enable compilation pipeline to go through the Yul intermediate representation

    Returns:
        build data dict
    """
    if not contract_sources:
        return {}
    if interface_sources is None:
        interface_sources = {}

    if [i for i in contract_sources if Path(i).suffix not in (".sol", ".vy")]:
        raise UnsupportedLanguage("Source suffixes must be one of ('.sol', '.vy')")
    if [i for i in interface_sources if Path(i).suffix not in (".sol", ".vy", ".json")]:
        raise UnsupportedLanguage("Interface suffixes must be one of ('.sol', '.vy', '.json')")

    build_json: dict[ContractName, ContractBuildJson] = {}
    compiler_targets = {}

    vyper_sources = {
        key: contract_sources[key] for key in contract_sources if Path(key).suffix == ".vy"
    }
    if vyper_sources:
        # TODO add `vyper_version` input arg to manually specify, support in config file
        if vyper_version is None:
            compiler_targets.update(
                find_vyper_versions(vyper_sources, install_needed=True, silent=silent)
            )
        else:
            compiler_targets[vyper_version] = list(vyper_sources)

    solc_sources = {
        key: contract_sources[key] for key in contract_sources if Path(key).suffix == ".sol"
    }
    if solc_sources:
        if solc_version is None:
            compiler_targets.update(
                find_solc_versions(solc_sources, install_needed=True, silent=silent)
            )
        else:
            compiler_targets[solc_version] = list(solc_sources)

        if optimizer is None:
            if optimize:
                optimizer = {"enabled": True, "runs": runs or 0}
            else:
                optimizer = {"enabled": False, "runs": 0}

    language: Language
    compiler_data: CompilerConfig
    for version, path_list in compiler_targets.items():
        compiler_data = {}  # type: ignore [typeddict-item]
        if path_list[0].endswith(".vy"):
            set_vyper_version(version)
            language = "Vyper"
            compiler_data["version"] = str(vyper.get_version())
            interfaces = {
                key: interface_sources[key]
                for key in interface_sources
                if Path(key).suffix != ".sol"
            }
        else:
            set_solc_version(version)
            language = "Solidity"
            compiler_data["version"] = str(solidity.get_version())
            interfaces = {
                k: v
                for k in interface_sources
                if Path(k).suffix == ".sol"
                and Version(version) in sources.get_pragma_spec(v := interface_sources[k], k)
            }

        to_compile = {key: contract_sources[key] for key in contract_sources if key in path_list}

        input_json = generate_input_json(
            to_compile,
            evm_version=evm_version[language] if isinstance(evm_version, dict) else evm_version,
            language=language,
            interface_sources=interfaces,
            remappings=remappings,
            optimizer=optimizer,
            viaIR=viaIR,
        )

        output_json = compile_from_input_json(input_json, silent, allow_paths)
        build_json.update(generate_build_json(input_json, output_json, compiler_data, silent))

    return build_json


def generate_input_json(
    contract_sources: dict[str, str],
    optimize: bool = True,
    runs: int = 200,
    evm_version: EvmVersion | None = None,
    language: Language = "Solidity",
    interface_sources: dict[str, str] | None = None,
    remappings: list[str] | str | None = None,
    optimizer: OptimizerSettings | None = None,
    viaIR: bool | None = None,
) -> InputJson:
    """Formats contracts to the standard solc input json.

    Args:
        contract_sources: a dictionary in the form of {path: 'source code'}
        optimize: (deprecated) enable solc optimizer
        runs: (deprecated) optimizer runs
        evm_version: evm version to compile for
        language: source language (Solidity or Vyper)
        interface_sources: dictionary of interfaces as {'path': "source code"}
        remappings: list of solidity path remappings
        optimizer: dictionary of solidity optimizer settings
        viaIR: enable compilation pipeline to go through the Yul intermediate representation

    Returns: dict
    """

    if language not in ("Solidity", "Vyper"):
        raise UnsupportedLanguage(f"{language}")

    if optimizer is None:
        optimizer = {"enabled": optimize, "runs": runs if optimize else 0}

    if evm_version is None:
        _module = solidity if language == "Solidity" else vyper
        evm_version = next(
            i[0] for i in _module.EVM_VERSION_MAPPING if _module.get_version() >= i[1]
        )

    input_json = deepcopy(STANDARD_JSON)
    input_json["language"] = language  # type: ignore [arg-type]
    settings = input_json["settings"]
    settings["evmVersion"] = evm_version
    if language == "Solidity":
        settings["optimizer"] = optimizer
        settings["remappings"] = _get_solc_remappings(remappings)
        if viaIR is not None:
            settings["viaIR"] = viaIR

    input_sources = _sources_dict(contract_sources, language)
    input_json["sources"] = input_sources

    if interface_sources:
        if language == "Solidity":
            input_sources.update(_sources_dict(interface_sources, language))
        else:
            input_json["interfaces"] = cast(
                dict[str, InterfaceSource],
                _sources_dict(interface_sources, language),
            )

    return input_json


def _get_solc_remappings(remappings: list[str] | str | None) -> list[str]:
    remap_dict: dict[str, str]
    if remappings is None:
        remap_dict = {}
    elif isinstance(remappings, str):
        remap_dict = dict([remappings.split("=")])
    else:
        remap_dict = dict(i.split("=") for i in remappings)
    remapped_dict: dict[str, str] = {}
    packages = _get_data_folder().joinpath("packages")
    for path in packages.iterdir():
        pathname = path.name
        key = next((k for k, v in remap_dict.items() if v.startswith(pathname)), None)
        if key:
            remapped_dict[key] = path.parent.joinpath(remap_dict.pop(key)).as_posix()
        else:
            remapped_dict[pathname] = path.as_posix()
    for k, v in remap_dict.items():
        if packages.joinpath(v).exists():
            remapped_dict[k] = packages.joinpath(v).as_posix()

    return [f"{k}={v}" for k, v in dict(remap_dict, **remapped_dict).items()]


def _get_allow_paths(allow_paths: str | None, remappings: list[str]) -> str:
    # generate the final allow_paths field based on path remappings
    path_list = [] if allow_paths is None else [allow_paths]

    remapping_paths = [i[i.index("=") + 1 :] for i in remappings]
    data_path = _get_data_folder().joinpath("packages").as_posix()
    remapping_paths = [i for i in remapping_paths if not i.startswith(data_path)]

    path_list = path_list + [data_path] + remapping_paths
    return ",".join(path_list)


def compile_from_input_json(
    input_json: InputJson, silent: bool = True, allow_paths: str | None = None
) -> dict:
    """
    Compiles contracts from a standard input json.

    Args:
        input_json: solc input json
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns: standard compiler output json
    """

    language = cast(Language, input_json["language"])
    if language == "Vyper":
        return vyper.compile_from_input_json(cast(InputJsonVyper, input_json), silent, allow_paths)

    if language == "Solidity":
        settings = cast(SettingsSolc, input_json["settings"])
        allow_paths = _get_allow_paths(allow_paths, settings["remappings"])
        return solidity.compile_from_input_json(
            cast(InputJsonSolc, input_json), silent, allow_paths
        )

    raise UnsupportedLanguage(language)


def generate_build_json(
    input_json: InputJson,
    output_json: dict,
    compiler_data: CompilerConfig | None = None,
    silent: bool = True,
) -> dict[ContractName, ContractBuildJson]:
    """Formats standard compiler output to the brownie build json.

    Args:
        input_json: solc input json used to compile
        output_json: output json returned by compiler
        compiler_data: additional data to include under 'compiler' in build json
        silent: verbose reporting

    Returns: build json dict"""

    language = input_json["language"]
    if language not in ("Solidity", "Vyper"):
        raise UnsupportedLanguage(language)

    if not silent:
        print("Generating build data...")

    compiler_data = compiler_data or {}  # type: ignore [assignment]

    settings = input_json["settings"]
    compiler_data["evm_version"] = settings["evmVersion"]
    build_json: dict[ContractName, ContractBuildJson] = {}

    if language == "Solidity":
        compiler_data["optimizer"] = settings["optimizer"]  # type: ignore [typeddict-item]
        source_nodes, statement_nodes, branch_nodes = solidity._get_nodes(output_json)

    sources = input_json["sources"]
    contracts: dict[str, dict[ContractName, dict]] = output_json["contracts"]

    for path_str, path_contracts in contracts.items():
        if path_str in sources:
            source: str = sources[path_str]["content"]  # type: ignore [typeddict-item]
            get_alias = False
        else:
            with Path(path_str).open(encoding="utf-8") as fp:
                source = fp.read()
            get_alias = True

        for contract_name, contract in path_contracts.items():
            if get_alias:
                contract_alias = _get_alias(contract_name, path_str)
            else:
                contract_alias = contract_name

            if not silent:
                print(f" - {contract_alias}")

            natspec = merge_natspec(contract.get("devdoc", {}), contract.get("userdoc", {}))

            abi: list[ABIElement] = contract["abi"]
            output_evm: dict = contract["evm"]
            deployed_bytecode: dict = output_evm["deployedBytecode"]
            bytecode: HexStr = deployed_bytecode["object"]

            if contract_alias in build_json and not bytecode:
                continue

            ast = output_json["sources"][path_str]["ast"]

            if language == "Solidity":
                contract_node = next(
                    i[contract_name] for i in source_nodes if i.absolutePath == path_str
                )
                build_json[contract_alias] = solidity._get_unique_build_json(
                    output_evm,
                    contract_node,
                    statement_nodes,
                    branch_nodes,
                    any(i["type"] == "fallback" for i in abi),
                )

            else:
                if contract_name == "<stdin>":
                    contract_name = contract_alias = ContractName("Vyper")
                build_json[contract_alias] = vyper._get_unique_build_json(
                    output_evm,
                    path_str,
                    contract_alias,
                    ast,
                    (0, len(source)),
                )

            build_json[contract_alias].update(
                {
                    "abi": abi,
                    "ast": ast,
                    "compiler": compiler_data,  # type: ignore [typeddict-item]
                    "contractName": contract_name,
                    "deployedBytecode": bytecode,
                    "deployedSourceMap": deployed_bytecode["sourceMap"],
                    "language": language,  # type: ignore [typeddict-item]
                    "natspec": natspec,
                    "opcodes": deployed_bytecode["opcodes"],
                    "sha1": HexStr(sha1(source.encode()).hexdigest()),
                    "source": source,
                    "sourceMap": output_evm["bytecode"].get("sourceMap", ""),
                    "sourcePath": path_str,
                }
            )
            size = len(bytecode.removeprefix("0x")) / 2
            if size > 24577:
                notify(
                    "WARNING",
                    f"deployed size of {contract_name} is {size} bytes, "
                    "exceeds EIP-170 limit of 24577",
                )

    if not silent:
        print("")

    return build_json


def _sources_dict(original: dict[str, Any], language: Language) -> SourcesDict:
    result: SourcesDict = {}
    for key, value in original.items():
        if Path(key).suffix == ".json":
            if isinstance(value, str):
                value = ujson_loads(value)
            result[key] = {"abi": value}
        else:
            result[key] = {"content": value}
    return result


def get_abi(
    contract_sources: dict[str, str],
    solc_version: str | None = None,
    allow_paths: str | None = None,
    remappings: list[str] | None = None,
    silent: bool = True,
) -> dict[ContractName, InterfaceBuildJson]:
    """
    Generate ABIs from contract interfaces.

    Arguments
    ---------
    contract_sources : dict
        a dictionary in the form of {'path': "source code"}
    solc_version: solc version to compile with (use None to set via pragmas)
    allow_paths : str, optional
        Compiler allowed filesystem import path
    remappings : list, optional
        List of solidity path remappings
    silent : bool, optional
        Disable verbose reporting

    Returns
    -------
    dict
        Compiled ABIs in the format `{'contractName': [ABI]}`
    """

    final_output: dict[ContractName, InterfaceBuildJson] = {
        ContractName(stem := p.stem): {
            "abi": ujson_loads(v),
            "contractName": ContractName(stem),
            "type": "interface",
            "source": None,
            "offset": None,
            "sha1": HexStr(sha1(v.encode()).hexdigest()),
        }
        for k, v in contract_sources.items()
        if (p := Path(k)).suffix == ".json"
    }

    for path, source in contract_sources.items():
        if Path(path).suffix == ".vy":
            input_json = generate_input_json({path: source}, language="Vyper")
            input_json["settings"]["outputSelection"]["*"] = {"*": ["abi"]}
            try:
                output_json = compile_from_input_json(input_json, silent, allow_paths)
            except Exception:
                # vyper interfaces do not convert to ABIs
                # https://github.com/vyperlang/vyper/issues/1944
                continue
            name = ContractName(Path(path).stem)
            final_output[name] = {
                "abi": output_json["contracts"][path][name]["abi"],
                "contractName": name,
                "type": "interface",
                "source": source,
                "offset": (0, len(source)),
                "sha1": HexStr(sha1(contract_sources[path].encode()).hexdigest()),
            }

    solc_sources = {k: v for k, v in contract_sources.items() if Path(k).suffix == ".sol"}

    if not solc_sources:
        return final_output

    if solc_version:
        compiler_targets = {solc_version: list(solc_sources)}
    else:
        compiler_targets = find_solc_versions(solc_sources, install_needed=True, silent=silent)

    for version, path_list in compiler_targets.items():
        to_compile = {k: v for k, v in contract_sources.items() if k in path_list}

        set_solc_version(version)
        input_json = generate_input_json(to_compile, language="Solidity", remappings=remappings)
        input_json["settings"]["outputSelection"]["*"] = {"*": ["abi"], "": ["ast"]}

        output_json = compile_from_input_json(input_json, silent, allow_paths)
        source_nodes = _from_standard_output(output_json)
        compiled_sources: dict[str, dict] = output_json["sources"]
        abi_json: dict[str, dict] = {
            k: v for k, v in output_json["contracts"].items() if k in path_list
        }

        for path, contracts in abi_json.items():
            path_source = contract_sources[path]
            for name, data in contracts.items():
                contract_node = next(i[name] for i in source_nodes if i.absolutePath == path)
                dependencies = []
                for node in contract_node.dependencies:
                    if node.nodeType == "ContractDefinition":
                        dependency_name = node.name
                        path_str = node.parent().absolutePath
                        dependencies.append(_get_alias(dependency_name, path_str))

                final_output[name] = {
                    "abi": data["abi"],
                    "ast": compiled_sources[path]["ast"],
                    "contractName": name,
                    "dependencies": dependencies,
                    "type": "interface",
                    "source": path_source,
                    "offset": contract_node.offset,
                    "sha1": HexStr(sha1(path_source.encode()).hexdigest()),
                }

    return final_output
